#define DIAGNOSTIC
/*
 * Copyright (c) 1991 Regents of the University of California.
 * All rights reserved.
 * Copyright (c) 1994,1998 John S. Dyson
 * All rights reserved.
 * Copyright (c) 1994 David Greenman
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department and William Jolitz of UUNET Technologies Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from:	@(#)pmap.c	7.7 (Berkeley)	5/12/91
 *	$Id: pmap.c,v 1.212 1998/11/08 02:26:14 msmith Exp $
 */

/*
 *	Manages physical address maps.
 *
 *	In addition to hardware address maps, this
 *	module is called upon to provide software-use-only
 *	maps which may or may not be stored in the same
 *	form as hardware maps.  These pseudo-maps are
 *	used to store intermediate results from copy
 *	operations to and from address spaces.
 *
 *	Since the information managed by this module is
 *	also stored by the logical address mapping module,
 *	this module may throw away valid virtual-to-physical
 *	mappings at almost any time.  However, invalidations
 *	of virtual-to-physical mappings must be done as
 *	requested.
 *
 *	In order to cope with hardware architectures which
 *	make virtual-to-physical map invalidates expensive,
 *	this module may delay invalidate or reduced protection
 *	operations until such time as they are actually
 *	necessary.  This module is given full information as
 *	to which processors are currently using which maps,
 *	and to when physical maps must be made correct.
 */

#include "opt_disable_pse.h"
#include "opt_pmap.h"
#include "opt_msgbuf.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/msgbuf.h>
#include <sys/vmmeter.h>
#include <sys/mman.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_prot.h>
#include <sys/lock.h>
#include <vm/vm_kern.h>
#include <vm/vm_page.h>
#include <vm/vm_map.h>
#include <vm/vm_object.h>
#include <vm/vm_extern.h>
#include <vm/vm_pageout.h>
#include <vm/vm_pager.h>
#include <vm/vm_zone.h>

#include <sys/user.h>

#include <machine/cputypes.h>
#include <machine/md_var.h>
#include <machine/specialreg.h>
#if defined(SMP) || defined(APIC_IO)
#include <machine/smp.h>
#include <machine/apic.h>
#endif /* SMP || APIC_IO */

#undef DISABLE_PSE
#define DISABLE_PSE

#define PMAP_KEEP_PDIRS
#ifndef PMAP_SHPGPERPROC
#define PMAP_SHPGPERPROC 200
#endif

#if defined(DIAGNOSTIC)
#define PMAP_DIAGNOSTIC
#endif

#define MINPV 2048

#if !defined(PMAP_DIAGNOSTIC)
#define PMAP_INLINE __inline
#else
#define PMAP_INLINE
#endif

/*
 * Get PDEs and PTEs for user/kernel address space
 */
#define	pmap_pde(m, v)	(&((m)->pm_pdir[(vm_offset_t)(v) >> PDRSHIFT]))
#define pdir_pde(m, v) (m[(vm_offset_t)(v) >> PDRSHIFT])

#define pmap_pde_v(pte)		((*(int *)pte & PG_V) != 0)
#define pmap_pte_w(pte)		((*(int *)pte & PG_W) != 0)
#define pmap_pte_m(pte)		((*(int *)pte & PG_M) != 0)
#define pmap_pte_u(pte)		((*(int *)pte & PG_A) != 0)
#define pmap_pte_v(pte)		((*(int *)pte & PG_V) != 0)

#define pmap_pte_set_w(pte, v) ((v)?(*(int *)pte |= PG_W):(*(int *)pte &= ~PG_W))
#define pmap_pte_set_prot(pte, v) ((*(int *)pte &= ~PG_PROT), (*(int *)pte |= (v)))

/*
 * Given a map and a machine independent protection code,
 * convert to a vax protection code.
 */
#define pte_prot(m, p)	(protection_codes[p])
static int protection_codes[8];

#define	pa_index(pa)		atop((pa) - vm_first_phys)
#define	pa_to_pvh(pa)		(&pv_table[pa_index(pa)])

static struct pmap kernel_pmap_store;
pmap_t kernel_pmap;
extern pd_entry_t my_idlePTD;

#if defined(SWTCH_OPTIM_STATS)
extern int tlb_flush_count;
#endif

vm_offset_t avail_start;	/* PA of first available physical page */
vm_offset_t avail_end;		/* PA of last available physical page */
vm_offset_t virtual_avail;	/* VA of first avail page (after kernel bss) */
vm_offset_t virtual_end;	/* VA of last avail page (end of kernel AS) */
static boolean_t pmap_initialized = FALSE;	/* Has pmap_init completed? */
static vm_offset_t vm_first_phys;
int pgeflag;		/* PG_G or-in */
int	staticpgeflag;	/* PG_G for static entities */
static int pseflag;		/* PG_PS or-in */
static int pv_npg;

static vm_object_t kptobj;

static int nkpt;
vm_offset_t kernel_vm_end;

/*
 * Data for the pv entry allocation mechanism
 */
static vm_zone_t pvzone;
static struct vm_zone pvzone_store;
static struct vm_object pvzone_obj;
static int pv_entry_count=0, pv_entry_max=0, pv_entry_high_water=0;
static int pmap_pagedaemon_waken = 0;
static struct pv_entry *pvinit;

/*
 * All those kernel PT submaps that BSD is so fond of
 */
pt_entry_t *CMAP1 = 0;
static pt_entry_t *CMAP2, *ptmmap;
static pv_table_t *pv_table;
caddr_t CADDR1 = 0, ptvmmap = 0;
static caddr_t CADDR2;
static pt_entry_t *msgbufmap;
struct msgbuf *msgbufp=0;

#ifdef SMP
extern char prv_CPAGE1[], prv_CPAGE2[], prv_CPAGE3[], prv_CPAGE4[];
extern pt_entry_t *prv_CMAP1, *prv_CMAP2, *prv_CMAP3, *prv_CMAP4;
extern pd_entry_t *IdlePTDS[];
extern pt_entry_t SMP_prvpt[];
extern pt_entry_t SMP_prvpte[];
extern volatile u_int cputlbwait[NCPU];
#endif

#ifdef SMP
extern unsigned int prv_PPAGE1[];
extern pt_entry_t *prv_PMAP1;
#else
static pt_entry_t *PMAP1 = 0;
static unsigned *PADDR1 = 0;
#endif

static PMAP_INLINE void	free_pv_entry __P((pv_entry_t pv));
static unsigned * get_ptbase __P((pmap_t pmap));
static pv_entry_t get_pv_entry __P((void));
static void	i386_protection_init __P((void));
static void	pmap_changebit __P((vm_page_t m, int bit, boolean_t setem));

static void	pmap_remove_all __P((vm_page_t m));
static vm_page_t pmap_enter_quick __P((pmap_t pmap, vm_offset_t va,
				      vm_page_t m, vm_page_t mpte));
static int pmap_remove_pte __P((struct pmap *pmap, unsigned *ptq,
					vm_offset_t sva, vm_page_t mptehint));
static void pmap_remove_page __P((struct pmap *pmap, vm_offset_t va));
static int pmap_remove_entry __P((struct pmap *pmap, unsigned oldpte,
					vm_offset_t va, vm_page_t mpte));
static boolean_t pmap_testbit __P((vm_page_t pa, int bit));
static void pmap_insert_entry __P((pmap_t pmap, vm_offset_t va,
		vm_page_t mpte, vm_page_t m));

static vm_page_t pmap_allocpte __P((pmap_t pmap, vm_offset_t va));

static int pmap_release_free_page __P((pmap_t pmap, vm_page_t p));
static vm_page_t _pmap_allocpte __P((pmap_t pmap, unsigned ptepindex));
static vm_page_t pmap_page_lookup __P((vm_object_t object, vm_pindex_t pindex));
static int pmap_unuse_pt __P((pmap_t, vm_offset_t, vm_page_t));
static vm_offset_t pmap_kmem_choose __P((vm_offset_t addr)) ;
void pmap_collect __P((void));
static void _pmap_zero_page __P((vm_offset_t va));
static vm_page_t pmap_getpte __P((pmap_t pmap, unsigned int ptepindex));
static void __inline page_wakeupq(vm_page_t *mpte);

static unsigned pdir4mb;

static void __inline
page_wakeupq(vm_page_t *mpte)
{
	vm_page_t m;
	if (mpte && (m = *mpte)) {
		vm_page_wakeup(m);
		*mpte = NULL;
	}
}

/*
 * Move the kernel virtual free pointer to the next
 * 4MB.  This is used to help improve performance
 * by using a large (4MB) page for much of the kernel
 * (.text, .data, .bss)
 */
static vm_offset_t
pmap_kmem_choose(vm_offset_t addr) {
	vm_offset_t newaddr = addr;
#ifndef DISABLE_PSE
	if (cpu_feature & CPUID_PSE) {
		newaddr = (addr + (NBPDR - 1)) & ~(NBPDR - 1);
	}
#endif
	return newaddr;
}

/*
 *	Bootstrap the system enough to run with virtual memory.
 *
 *	On the i386 this is called after mapping has already been enabled
 *	and just syncs the pmap module with what has already been done.
 *	[We can't call it easily with mapping off since the kernel is not
 *	mapped with PA == VA, hence we would have to relocate every address
 *	from the linked base (virtual) address "KERNBASE" to the actual
 *	(physical) address starting relative to 0]
 */
void
pmap_bootstrap(firstaddr, loadaddr)
	vm_offset_t firstaddr;
	vm_offset_t loadaddr;
{
	vm_offset_t va;
	pt_entry_t *pte;
	int i, j;

	avail_start = firstaddr;

	/*
	 * XXX The calculation of virtual_avail is wrong. It's NKPT*PAGE_SIZE too
	 * large. It should instead be correctly calculated in locore.s and
	 * not based on 'first' (which is a physical address, not a virtual
	 * address, for the start of unused physical memory). The kernel
	 * page tables are NOT double mapped and thus should not be included
	 * in this calculation.
	 */
	virtual_avail = (vm_offset_t) KERNBASE + firstaddr;
	virtual_avail = pmap_kmem_choose(virtual_avail);

	virtual_end = VM_MAX_KERNEL_ADDRESS;

	/*
	 * Initialize protection array.
	 */
	i386_protection_init();

	/*
	 * The kernel's pmap is statically allocated so we don't have to use
	 * pmap_create, which is unlikely to work correctly at this part of
	 * the boot sequence (XXX and which no longer exists).
	 */
	kernel_pmap = &kernel_pmap_store;

	kernel_pmap->pm_pdir = (pd_entry_t *) (KERNBASE + (u_int)IdlePTD);
	kernel_pmap->pm_pactive = 0xffffffff;

	kernel_pmap->pm_count = 1;
	TAILQ_INIT(&kernel_pmap->pm_pvlist);
	nkpt = NKPT;

	/*
	 * Reserve some special page table entries/VA space for temporary
	 * mapping of pages.
	 */
#define	SYSMAP(c, p, v, n)	\
	v = (c)va; va += ((n)*PAGE_SIZE); p = pte; pte += (n);

	va = virtual_avail;
	pte = (pt_entry_t *) pmap_pte(kernel_pmap, va);

	/*
	 * CMAP1/CMAP2 are used for zeroing and copying pages.
	 */
	SYSMAP(caddr_t, CMAP1, CADDR1, 1)
	SYSMAP(caddr_t, CMAP2, CADDR2, 1)

	/*
	 * ptvmmap is used for reading arbitrary physical pages via /dev/mem.
	 * XXX ptmmap is not used.
	 */
	SYSMAP(caddr_t, ptmmap, ptvmmap, 1)

	/*
	 * msgbufp is used to map the system message buffer.
	 * XXX msgbufmap is not used.
	 */
	SYSMAP(struct msgbuf *, msgbufmap, msgbufp,
	       atop(round_page(MSGBUF_SIZE)))

#if !defined(SMP)
	/*
	 * ptemap is used for pmap_pte_quick
	 */
	SYSMAP(unsigned *, PMAP1, PADDR1, 1);
#endif

	virtual_avail = va;

	*(int *) CMAP1 = *(int *) CMAP2 = 0;
	*(int *) PTD = 0;


	pgeflag = 0;
	staticpgeflag = 0;
	if (cpu_feature & CPUID_PGE) {
#if !defined(SMP)
		pgeflag = PG_G;
#endif
		staticpgeflag = PG_G;
	}
	
/*
 * Initialize the 4MB page size flag
 */
	pseflag = 0;
/*
 * The 4MB page version of the initial
 * kernel page mapping.
 */
	pdir4mb = 0;

#if !defined(DISABLE_PSE)
	if (cpu_feature & CPUID_PSE) {
		unsigned ptditmp;
		/*
		 * Enable the PSE mode
		 */
		load_cr4(rcr4() | CR4_PSE);

		/*
		 * Note that we have enabled PSE mode
		 */
		pseflag = PG_PS;
		ptditmp = *((unsigned *)PTmap + i386_btop(KERNBASE));
		ptditmp &= ~(NBPDR - 1);
		ptditmp |= PG_V | PG_RW | PG_PS | PG_A | PG_U | staticpgeflag;
		pdir4mb = ptditmp;
		/*
		 * We can do the mapping here for the single processor
		 * case.  We simply ignore the old page table page from
		 * now on.
		 */
#if !defined(SMP)
		setpte(&PTD[KPTDI], ptditmp);
		setpte(&kernel_pmap->pm_pdir[KPTDI], ptditmp);
		invltlb();
#endif
	}
#endif

#ifdef SMP
	if (cpu_apic_address == 0)
		panic("pmap_bootstrap: no local apic!");

	/* 0 = private page */
	/* 1 = page table page */
	/* 2 = local apic */
	/* 16-31 = io apics */
	setpte(&SMP_prvpt[2], PG_V | PG_RW |
		PG_A | PG_M | staticpgeflag | (cpu_apic_address & PG_FRAME));

	for (i = 0; i < mp_napics; i++) {
		for (j = 0; j < 16; j++) {
			/* same page frame as a previous IO apic? */
			if (((vm_offset_t)SMP_prvpt[j + 16] & PG_FRAME) ==
			    (io_apic_address[0] & PG_FRAME)) {
				ioapic[i] = (ioapic_t *)&SMP_ioapic[j * PAGE_SIZE];
				break;
			}
			/* use this slot if available */
			if (((vm_offset_t)SMP_prvpt[j + 16] & PG_FRAME) == 0) {
				setpte(&SMP_prvpt[j + 16], PG_V | PG_RW | PG_A | PG_M |
				    (io_apic_address[i] & PG_FRAME));
				ioapic[i] = (ioapic_t *)&SMP_ioapic[j * PAGE_SIZE];
				break;
			}
		}
		if (j == 16)
			panic("no space to map IO apic %d!", i);
	}

	/* BSP does this itself, AP's get it pre-set */
	prv_CMAP1 = &SMP_prvpt[3 + UPAGES];
	prv_CMAP2 = &SMP_prvpt[4 + UPAGES];
	prv_CMAP3 = &SMP_prvpt[5 + UPAGES];
	prv_CMAP4 = &SMP_prvpt[6 + UPAGES];
	prv_PMAP1 = &SMP_prvpt[7 + UPAGES];
#endif

	invltlb();

}

void
getmtrr()
{
	int i;

	if (cpu_class == CPUCLASS_686) {
		for(i = 0; i < NPPROVMTRR; i++) {
			PPro_vmtrr[i].base = rdmsr(PPRO_VMTRRphysBase0 + i * 2);
			PPro_vmtrr[i].mask = rdmsr(PPRO_VMTRRphysMask0 + i * 2);
		}
	}
}

void
putmtrr()
{
	int i;

	if (cpu_class == CPUCLASS_686) {
		wbinvd();
		for(i = 0; i < NPPROVMTRR; i++) {
			wrmsr(PPRO_VMTRRphysBase0 + i * 2, PPro_vmtrr[i].base);
			wrmsr(PPRO_VMTRRphysMask0 + i * 2, PPro_vmtrr[i].mask);
		}
	}
}

void
pmap_setvidram(void)
{
#if 0
	if (cpu_class == CPUCLASS_686) {
		wbinvd();
		/*
		 * Set memory between 0-640K to be WB
		 */
		wrmsr(0x250, 0x0606060606060606LL);
		wrmsr(0x258, 0x0606060606060606LL);
		/*
		 * Set normal, PC video memory to be WC
		 */
		wrmsr(0x259, 0x0101010101010101LL);
	}
#endif
}

void
pmap_setdevram(unsigned long long basea, vm_offset_t sizea)
{
	int i, free, skip;
	unsigned basepage, basepaget;
	unsigned long long base;
	unsigned long long mask;

	if (cpu_class != CPUCLASS_686)
		return;

	free = -1;
	skip = 0;
	basea &= ~0xfff;
	base = basea | 0x1;
	mask = (long long) (0xfffffffffLL - ((long) sizea - 1)) | (long long) 0x800;
	mask &= ~0x7ff;

	basepage = (long long) (base >> 12);
	for(i = 0; i < NPPROVMTRR; i++) {
		PPro_vmtrr[i].base = rdmsr(PPRO_VMTRRphysBase0 + i * 2);
		PPro_vmtrr[i].mask = rdmsr(PPRO_VMTRRphysMask0 + i * 2);
		basepaget = (long long) (PPro_vmtrr[i].base >> 12);
		if (basepage == basepaget)
			skip = 1;
		if ((PPro_vmtrr[i].mask & 0x800) == 0) {
			if (free == -1)
				free = i;
		}
	}

	if (!skip && free != -1) {
		wbinvd();
		PPro_vmtrr[free].base = base;
		PPro_vmtrr[free].mask = mask;
		wrmsr(PPRO_VMTRRphysBase0 + free * 2, base);
		wrmsr(PPRO_VMTRRphysMask0 + free * 2, mask);
		printf(
	"pmap: added WC mapping at page: 0x%x %x, size: %u mask: 0x%x %x\n",
		    (u_int)(base >> 32), (u_int)base, sizea,
		    (u_int)(mask >> 32), (u_int)mask);
	}
}

/*
 * Set 4mb pdir for mp startup, and global flags
 */
void
pmap_set_opt(unsigned *pdir) {
	int i;

	if (pseflag) {
		load_cr4(rcr4() | CR4_PSE);
		if (pdir4mb) {
			pdir[KPTDI] = pdir4mb;
		}
	}

	if (staticpgeflag) {
		load_cr4(rcr4() | CR4_PGE);
		for(i = KPTDI; i < KPTDI + nkpt; i++) {
			if (pdir[i]) {
				pdir[i] |= PG_G;
			}
		}
	}
}

/*
 * Setup the PTD for the boot processor
 */
void
pmap_set_opt_bsp(void)
{
	pmap_set_opt((unsigned *)kernel_pmap->pm_pdir);
	pmap_set_opt((unsigned *)PTD);
	invltlb();
}

/*
 *	Initialize the pmap module.
 *	Called by vm_init, to initialize any structures that the pmap
 *	system needs to map virtual memory.
 *	pmap_init has been enhanced to support in a fairly consistant
 *	way, discontiguous physical memory.
 */
void
pmap_init(phys_start, phys_end)
	vm_offset_t phys_start, phys_end;
{
	vm_offset_t addr;
	vm_size_t s;
	int i;
	int initial_pvs;

	/*
	 * calculate the number of pv_entries needed
	 */
	vm_first_phys = phys_avail[0];
	for (i = 0; phys_avail[i + 1]; i += 2);
	pv_npg = (phys_avail[(i - 2) + 1] - vm_first_phys) / PAGE_SIZE;

	/*
	 * Allocate memory for random pmap data structures.  Includes the
	 * pv_head_table.
	 */
	s = (vm_size_t) (sizeof(pv_table_t) * pv_npg);
	s = round_page(s);

	addr = (vm_offset_t) kmem_alloc(kernel_map, s);
	pv_table = (pv_table_t *) addr;
	for(i = 0; i < pv_npg; i++) {
		vm_offset_t pa;
		vm_page_t m;
		TAILQ_INIT(&pv_table[i].pv_list);
		pv_table[i].pv_list_count = 0;
		pa = vm_first_phys + i * PAGE_SIZE;
		m = PHYS_TO_VM_PAGE(pa);
		m->pventry = &pv_table[i];
		pv_table[i].pv_vm_page = m;
	}

	/*
	 * init the pv free list
	 */
	initial_pvs = pv_npg;
	if (initial_pvs < MINPV)
		initial_pvs = MINPV;
	pvzone = &pvzone_store;
	pvinit = (struct pv_entry *) kmem_alloc(kernel_map,
		initial_pvs * sizeof (struct pv_entry));
	zbootinit(pvzone, "PV ENTRY", sizeof (struct pv_entry), pvinit, pv_npg);
	/*
	 * object for kernel page table pages
	 */
	if (kptobj == NULL)
		kptobj = vm_object_allocate(OBJT_DEFAULT, NKPDE);

	/*
	 * Now it is safe to enable pv_table recording.
	 */
	pmap_initialized = TRUE;
}

/*
 * Initialize the address space (zone) for the pv_entries.  Set a
 * high water mark so that the system can recover from excessive
 * numbers of pv entries.
 */
void
pmap_init2() {
	pv_entry_max = PMAP_SHPGPERPROC * maxproc + pv_npg;
	pv_entry_high_water = 9 * (pv_entry_max / 10);
	zinitna(pvzone, &pvzone_obj, NULL, 0, pv_entry_max, ZONE_INTERRUPT, 1);
}

/*
 *	Used to map a range of physical addresses into kernel
 *	virtual address space.
 *
 *	For now, VM is already on, we only need to map the
 *	specified memory.
 */
vm_offset_t
pmap_map(virt, start, end, prot)
	vm_offset_t virt;
	vm_offset_t start;
	vm_offset_t end;
	int prot;
{
	while (start < end) {
		pmap_kenterp(virt, start);
		virt += PAGE_SIZE;
		start += PAGE_SIZE;
	}
	return (virt);
}


/***************************************************
 * Low level helper routines.....
 ***************************************************/

#if defined(PMAP_DIAGNOSTIC)

/*
 * This code checks for non-writeable/modified pages.
 * This should be an invalid condition.
 */
static int
pmap_nw_modified(pt_entry_t ptea) {
	int pte;

	pte = (int) ptea;

	if ((pte & (PG_M|PG_RW)) == PG_M)
		return 1;
	else
		return 0;
}
#endif


/*
 * this routine defines the region(s) of memory that should
 * not be tested for the modified bit.
 */
static PMAP_INLINE int
pmap_track_modified( vm_offset_t va) {
	if ((va < clean_sva) || (va >= clean_eva)) 
		return 1;
	else
		return 0;
}

/*
 * TLB invalidation routines
 */
static __inline void
cpu_invlpg(u_int addr)
{
	__asm	__volatile("invlpg %0"::"m"(*(char *)addr):"memory");
}

static __inline void
cpu_invltlb(void)
{
	u_int	temp;
	/*
	 * This should be implemented as load_cr3(rcr3()) when load_cr3()
	 * is inlined.
	 */
	__asm __volatile("movl %%cr3, %0; movl %0, %%cr3" : "=r" (temp)
			 : : "memory");
#ifdef SMP
	cputlbwait[cpuid] = 0;
#endif
#if defined(SWTCH_OPTIM_STATS)
	++tlb_flush_count;
#endif
}

void
invlpg(u_int addr)
{   
#if defined(I386_CPU)
	if (cpu_class == CPUCLASS_386) {
		invltlb();
	} else
#endif
	{
		cpu_invlpg(addr);
	}
#ifdef SMP
	/* send a message to the other CPUs */
	smp_invltlb();
#endif
}

void
invltlb(void)
{
	cpu_invltlb();
#ifdef SMP
	/* send a message to the other CPUs */
	smp_invltlb();
#endif
}

#if defined(SMP)
static void
invlpgcpu(vm_offset_t va, int mask)
{
	if (mask & ~(1 << cpuid))
		invlpg(va);
	else
		cpu_invlpg(va);
}

static void
invltlbcpu(int mask)
{
	int temp;
	if (mask == 0)
		return;
	if (mask & ~(1 << cpuid))
		invltlb();
	else
		cpu_invltlb();
}
#endif

static void 
_pmap_zero_page(vm_offset_t va)
{
#if defined(I686_CPU)
	if (cpu == CPU_686) {
		i686_pagezero((caddr_t)va);
	} else
#endif
		bzero((caddr_t) va, PAGE_SIZE);
}

static unsigned *
get_ptbase(pmap)
	pmap_t pmap;
{
	unsigned frame = (unsigned) pmap->pm_pdir[PTDPTDI] & PG_FRAME;

	/* are we current address space or kernel? */
	if (pmap == kernel_pmap || frame == (((unsigned) PTDpde) & PG_FRAME)) {
		return (unsigned *) PTmap;
	}
	/* otherwise, we are alternate address space */
	if (frame != (((unsigned) APTDpde) & PG_FRAME)) {
		setpte(&APTDpde, frame | PG_RW | PG_V | PG_A);
#if defined(SMP)
		/* The page directory is not shared between CPUs */
		cpu_invltlb();
#else
		invltlb();
#endif
	}
	return (unsigned *) APTmap;
}

/*
 * Super fast pmap_pte routine best used when scanning
 * the pv lists.  This eliminates many coarse-grained
 * invltlb calls.  Note that many of the pv list
 * scans are across different pmaps.  It is very wasteful
 * to do an entire invltlb for checking a single mapping.
 */

unsigned * 
pmap_pte(pmap, va)
	register pmap_t pmap;
	vm_offset_t va;
{
	unsigned pde, newpf;
	unsigned *pdep;
	unsigned pdirindex = va >> PDRSHIFT;
	if (pde = (unsigned) pmap->pm_pdir[pdirindex]) {
		unsigned frame;
		unsigned index;

		if (pde & PG_PS) {
			return (unsigned *) &pmap->pm_pdir[pdirindex];
		}

		frame = ((unsigned) pmap->pm_pdir[PTDPTDI]) & PG_FRAME;

		index = i386_btop(va);
		/* are we current address space or kernel? */

		if (frame == (((unsigned) PTDpde) & PG_FRAME))
			return (unsigned *) PTmap + index;

		if (pmap == kernel_pmap)
			return (unsigned *) PTmap + index;

		newpf = pde & PG_FRAME;

#ifdef SMP
		if ( ((* (unsigned *) prv_PMAP1) & PG_FRAME) != newpf) {
			setpte(prv_PMAP1, newpf | PG_RW | PG_V | PG_A | PG_M);
			cpu_invlpg((vm_offset_t)&prv_PPAGE1);
		}
		return prv_PPAGE1 + ((unsigned) index & (NPTEPG - 1));
#else
		if ( ((* (unsigned *) PMAP1) & PG_FRAME) != newpf) {
			setpte(PMAP1, newpf | PG_RW | PG_V | PG_A | PG_M);
			invlpg((vm_offset_t) PADDR1);
		}
		return PADDR1 + ((unsigned) index & (NPTEPG - 1));
#endif
	}
	return (0);
}

vm_page_t
pmap_getpte(pmap_t pmap, unsigned int ptepindex)
{
	vm_page_t mpte, mptehint;
retry:
	if ((mptehint = pmap->pm_ptphint) &&
		(mptehint->object == pmap->pm_pteobj) &&
		(mptehint->pindex == ptepindex)) {
		mpte = mptehint;
		if (vm_page_sleep(mpte,"pmpunu",NULL))
			goto retry;
		vm_page_busy(mpte);
	} else {
		mpte = pmap_page_lookup( pmap->pm_pteobj, ptepindex);
		if (mpte == NULL)
			return NULL;
		pmap->pm_ptphint = mpte;
	}
	return mpte;
}

/*
 *	Routine:	pmap_extract
 *	Function:
 *		Extract the physical page address associated
 *		with the given map/virtual_address pair.
 */
vm_offset_t 
pmap_extract(pmap, va)
	register pmap_t pmap;
	vm_offset_t va;
{
	unsigned pde, newpf;
	unsigned *pte;
	unsigned index;
	unsigned rtval;
	if (pmap && (pde = (unsigned) pmap->pm_pdir[va >> PDRSHIFT])) {
		newpf = pde & PG_FRAME;
		if ((pde & PG_PS) != 0) {
			pde &= ~(NBPDR - 1);
			pde |= va & (NBPDR - 1);
			return pde;
		}
		index = i386_btop(va);
#ifdef SMP
		if ( ((* (unsigned *) prv_PMAP1) & PG_FRAME) != newpf) {
			setpte(prv_PMAP1, newpf | PG_RW | PG_V | PG_A | PG_M);
			cpu_invlpg((vm_offset_t)&prv_PPAGE1);
		}
		pte = prv_PPAGE1 + ((unsigned) index & (NPTEPG - 1));
#else
		if ( ((* (unsigned *) PMAP1) & PG_FRAME) != newpf) {
			setpte(PMAP1, newpf | PG_RW | PG_V | PG_A | PG_M);
			invlpg((vm_offset_t) PADDR1);
		}
		pte = PADDR1 + ((unsigned) index & (NPTEPG - 1));
#endif
		return *pte | (va & PAGE_MASK);
	}
	return 0;

}

/***************************************************
 * Low level mapping routines.....
 ***************************************************/

/*
 * Add a list of wired pages to the kva
 * this routine is only used for temporary
 * kernel mappings that do not need to have
 * page modification or references recorded.
 * Note that old mappings are simply written
 * over.  The page *must* be wired.
 */
void
pmap_qenter(va, m, count)
	vm_offset_t va;
	vm_page_t *m;
	int count;
{
	int i;
	unsigned *ptep;
	unsigned lobits;

	lobits = PG_RW | PG_V | PG_A | PG_M | pgeflag;
	ptep = (unsigned *)vtopte(va);

	for (i = 0; i < count; i++, va += PAGE_SIZE) {
		unsigned npte, opte;
		opte = setpte(&ptep[i], VM_PAGE_TO_PHYS(m[i])|lobits);
		if (opte & PG_V)
			invlpg(va);
	}
}

/*
 * this routine jerks page mappings from the
 * kernel -- it is meant only for temporary mappings.
 */
void
pmap_qremove(va, count)
	vm_offset_t va;
	int count;
{
	int i;
	register unsigned *ptep;

	ptep = (unsigned *)vtopte(va);
	for (i = 0; i < count; i++, va += PAGE_SIZE) {
		setpte(&ptep[i], 0);
		invlpg(va);
	}
}

/*
 * add a wired page to the kva
 * note that in order for the mapping to take effect -- you
 * should do a invltlb after doing the pmap_kenter...
 */
void 
pmap_kenter(va, pa)
	vm_offset_t va;
	register vm_offset_t pa;
{
	register unsigned *pte;
	unsigned npte, opte;

	pte = (unsigned *)vtopte(va);
	npte = pa | PG_RW | PG_V | PG_A | PG_M | pgeflag;
	if (setpte(pte, npte))
		invlpg(va);
}

/*
 * like pmap_kenter, but has PG_G even on SMP machines.  This mapping
 * requires an explicit invlpg to uncache it in the TLB.
 */
void 
pmap_kenterp(va, pa)
	vm_offset_t va, pa;
{
	register unsigned *pte;
	unsigned npte, opte;

	pte = (unsigned *)vtopte(va);
#if defined(SMP)
	npte = pa | PG_RW | PG_V | PG_A | PG_M | staticpgeflag;
#else
	npte = pa | PG_RW | PG_V | PG_A | PG_M | pgeflag;
#endif
	if (setpte(pte, npte))
		invlpg(va);
}

/*
 * remove a page from the kernel pagetables
 */
void
pmap_kremove(va)
	vm_offset_t va;
{
	register unsigned *pte;

	pte = (unsigned *)vtopte(va);
	setpte(pte, 0);
	invlpg(va);
}

static vm_page_t
pmap_page_lookup(object, pindex)
	vm_object_t object;
	vm_pindex_t pindex;
{
	vm_page_t m;
retry:
	m = vm_page_lookup(object, pindex);
	if (m == NULL)
		return NULL;
	if (vm_page_sleep(m, "pplookp", 0))
		goto retry;
	vm_page_busy(m);
	return m;
}

/*
 * Create the UPAGES for a new process.
 * This routine directly affects the fork perf for a process.
 */
void
pmap_new_proc(p)
	struct proc *p;
{
	int i, updateneeded, immediatelyvalid;
	vm_object_t upobj;
	vm_page_t m;
	struct user *up;
	unsigned *ptek, oldpte;

	/*
	 * allocate object for the upages
	 */
	if ((upobj = p->p_upages_obj) == NULL) {
		upobj = vm_object_allocate( OBJT_DEFAULT, UPAGES);
		p->p_upages_obj = upobj;
	}

	/* get a kernel virtual address for the UPAGES for this proc */
	if ((up = p->p_addr) == NULL) {
		up = (struct user *) kmem_alloc_pageable(kernel_map,
				UPAGES * PAGE_SIZE);
#if !defined(MAX_PERF)
		if (up == NULL)
			panic("pmap_new_proc: u_map allocation failed");
#endif
		p->p_addr = up;
	}

	ptek = (unsigned *) vtopte((vm_offset_t) up);

	updateneeded = 0;
	for(i=0;i<UPAGES;i++) {
		/*
		 * Get a kernel stack page
		 */
		m = vm_page_grab(upobj, i, VM_ALLOC_NORMAL | VM_ALLOC_RETRY);

		/*
		 * Wire the page
		 */
		m->wire_count++;
		cnt.v_wire_count++;

		/*
		 * Enter the page into the kernel address space.
		 */
		oldpte = setpte(ptek + i,
			VM_PAGE_TO_PHYS(m) | PG_RW | PG_V | PG_A | PG_M | pgeflag);
#if defined(I386_CPU)
		if (oldpte) {
			if (cpu_class > CPUCLASS_386) {
				invlpg((vm_offset_t) up + i * PAGE_SIZE);
				immediatelyvalid = 1;
			} else {
				updateneeded = 1;
				immediatelyvalid = 0;
			}
		} else {
			immediatelyvalid = 1;
		}
#else
#if !defined(SMP)
		if (oldpte) {
			invlpg((vm_offset_t) up + i * PAGE_SIZE);
		}
#else
		if (oldpte) {
			updateneeded = 1;
		}
#endif
#endif
		vm_page_wakeup(m);
		m->flags &= ~PG_ZERO;
		m->flags |= PG_MAPPED | PG_WRITEABLE;
		m->valid = VM_PAGE_BITS_ALL;
	}

#if defined(I386_CPU) || defined(SMP)
	if (updateneeded)
		invltlb();
#endif
}

/*
 * Dispose the UPAGES for a process that has exited.
 * This routine directly impacts the exit perf of a process.
 */
void
pmap_dispose_proc(p)
	struct proc *p;
{
	int i;
	vm_object_t upobj;
	vm_page_t m;
	unsigned *ptek, oldpte;

	upobj = p->p_upages_obj;

	oldpte = 0;
	ptek = (unsigned *) vtopte((vm_offset_t) p->p_addr);
	for(i=0;i<UPAGES;i++) {

		if ((m = vm_page_lookup(upobj, i)) == NULL)
			panic("pmap_dispose_proc: upage already missing???");

		vm_page_busy(m);

		oldpte |= setpte(ptek + i, 0);
#if !defined(SMP)
		if (oldpte && (cpu_class > CPUCLASS_386))
			invlpg((vm_offset_t) p->p_addr + i * PAGE_SIZE);
#endif
		vm_page_unwire(m, 0);
		vm_page_free(m);
	}

#if !defined(SMP)
	if (oldpte && (cpu_class <= CPUCLASS_386))
#endif
		invltlb();
}

/*
 * Allow the UPAGES for a process to be prejudicially paged out.
 */
void
pmap_swapout_proc(p)
	struct proc *p;
{
	int i;
	vm_object_t upobj;
	vm_page_t m;

	upobj = p->p_upages_obj;
	/*
	 * let the upages be paged
	 */
	for(i=0;i<UPAGES;i++) {
		if ((m = vm_page_lookup(upobj, i)) == NULL)
			panic("pmap_swapout_proc: upage already missing???");
		m->dirty = VM_PAGE_BITS_ALL;
		vm_page_unwire(m, 0);
		pmap_kremove( (vm_offset_t) p->p_addr + PAGE_SIZE * i);
	}
}

/*
 * Bring the UPAGES for a specified process back in.
 */
void
pmap_swapin_proc(p)
	struct proc *p;
{
	int i,rv;
	vm_object_t upobj;
	vm_page_t m;

	upobj = p->p_upages_obj;
	for(i=0;i<UPAGES;i++) {

		m = vm_page_grab(upobj, i, VM_ALLOC_NORMAL | VM_ALLOC_RETRY);

		pmap_kenter(((vm_offset_t) p->p_addr) + i * PAGE_SIZE,
			VM_PAGE_TO_PHYS(m));

		if (m->valid != VM_PAGE_BITS_ALL) {
			rv = vm_pager_get_pages(upobj, &m, 1, 0);
#if !defined(MAX_PERF)
			if (rv != VM_PAGER_OK)
				panic("pmap_swapin_proc: cannot get upages for proc: %d\n", p->p_pid);
#endif
			m = vm_page_lookup(upobj, i);
			m->valid = VM_PAGE_BITS_ALL;
		}

		vm_page_wire(m);
		vm_page_wakeup(m);
		m->flags |= PG_MAPPED | PG_WRITEABLE;
	}
}

/***************************************************
 * Page table page management routines.....
 ***************************************************/

/*
 * This routine unholds page table pages, and if the hold count
 * drops to zero, then it decrements the wire count.
 */
static int 
_pmap_unwire_pte_hold(pmap_t pmap, vm_page_t m) {
	int s;

	vm_offset_t pteva;
	/*
	 * unmap the page table page
	 */
	pmap->pm_pdir[m->pindex] = 0;
	--pmap->pm_stats.resident_count;
	if (pmap->pm_pactive) {
		/*
		 * Do a invltlb to make the invalidated mapping
		 * take effect immediately.
		 */
		pteva = UPT_MIN_ADDRESS + i386_ptob(m->pindex);
#if defined(SMP)
		invlpgcpu(pteva, pmap->pm_pactive);
#else
		invlpg(pteva);
#endif
	}

	if (pmap->pm_ptphint == m)
		pmap->pm_ptphint = NULL;

	/*
	 * If the page is finally unwired, simply free it.
	 */
	--m->wire_count;
	if (m->wire_count == 0) {
		vm_page_free_zero(m);
		--cnt.v_wire_count;
		return 1;
	}
	return 0;
}

static PMAP_INLINE int
pmap_unwire_pte_hold(pmap_t pmap, vm_page_t m) {
	vm_page_unhold(m);
	if (m->hold_count == 0)
		return _pmap_unwire_pte_hold(pmap, m);
	return 0;
}

/*
 * After removing a page table entry, this routine is used to
 * conditionally free the page, and manage the hold/wire counts.
 */
static int
pmap_unuse_pt(pmap, va, mpte)
	pmap_t pmap;
	vm_offset_t va;
	vm_page_t mpte;
{
	int rtval;

	if ((va >= UPT_MIN_ADDRESS) || (pmap == kernel_pmap))
		return 0;

	if (mpte == NULL) {
		mpte = pmap_getpte(pmap, (va >> PDRSHIFT));
		if (mpte && !pmap_unwire_pte_hold(pmap, mpte)) {
			vm_page_wakeup(mpte);
			return 0;
		}
		return 0;
	}
	return pmap_unwire_pte_hold(pmap, mpte);
}

#if !defined(SMP)
void
pmap_pinit0(pmap)
	struct pmap *pmap;
{
	pmap->pm_pdir =
		(pd_entry_t *)kmem_alloc_pageable(kernel_map, PAGE_SIZE);
	pmap_kenter((vm_offset_t) pmap->pm_pdir, (vm_offset_t) IdlePTD);
	pmap->pm_flags = 0;
	pmap->pm_count = 1;
	pmap->pm_ptphint = NULL;
	pmap->pm_pactive = 0;
	TAILQ_INIT(&pmap->pm_pvlist);
	bzero(&pmap->pm_stats, sizeof pmap->pm_stats);
}
#else
void
pmap_pinit0(pmap)
	struct pmap *pmap;
{
	pmap_pinit(pmap);
}
#endif

/*
 * Initialize a preallocated and zeroed pmap structure,
 * such as one in a vmspace structure.
 */
void
pmap_pinit(pmap)
	register struct pmap *pmap;
{
	vm_page_t ptdpg;

	/*
	 * No need to allocate page table space yet but we do need a valid
	 * page directory table.
	 */
	if (pmap->pm_pdir == NULL)
		pmap->pm_pdir =
			(pd_entry_t *)kmem_alloc_pageable(kernel_map, PAGE_SIZE);

	/*
	 * allocate object for the ptes
	 */
	if (pmap->pm_pteobj == NULL)
		pmap->pm_pteobj = vm_object_allocate( OBJT_DEFAULT, PTDPTDI + 1);

	/*
	 * allocate the page directory page
	 */
retry:
	ptdpg = vm_page_grab( pmap->pm_pteobj, PTDPTDI,
			VM_ALLOC_ZERO | VM_ALLOC_RETRY);

	ptdpg->wire_count = 1;
	++cnt.v_wire_count;

	ptdpg->flags &= ~PG_MAPPED;	/* not mapped normally */
	vm_page_wakeup(ptdpg);
	ptdpg->valid = VM_PAGE_BITS_ALL;

	pmap_kenter((vm_offset_t) pmap->pm_pdir, VM_PAGE_TO_PHYS(ptdpg));
	if ((ptdpg->flags & PG_ZERO) == 0)
		_pmap_zero_page((vm_offset_t)pmap->pm_pdir);

	/* wire in kernel global address entries */
	/* XXX copies current process, does not fill in MPPTDI */
	bcopy(PTD + KPTDI, pmap->pm_pdir + KPTDI, nkpt * PTESIZE);
#if defined(SMP)
	pmap->pm_pdir[MPPTDI] = SMP_prvpte[cpuid];
#endif
	pmap->pm_pdir[APTDPTDI] = 0;

	/* install self-referential address mapping entry */
	setpte(&pmap->pm_pdir[PTDPTDI],
		(VM_PAGE_TO_PHYS(ptdpg) | PG_V | PG_RW | PG_A | PG_M));

	pmap->pm_flags = 0;
	pmap->pm_count = 1;
	pmap->pm_ptphint = NULL;
	pmap->pm_pactive = 0;
	TAILQ_INIT(&pmap->pm_pvlist);
	bzero(&pmap->pm_stats, sizeof pmap->pm_stats);
}

static int
pmap_release_free_page(pmap, p)
	struct pmap *pmap;
	vm_page_t p;
{
	int s;
	unsigned *pde = (unsigned *) pmap->pm_pdir;
	/*
	 * This code optimizes the case of freeing non-busy
	 * page-table pages.  Those pages are zero now, and
	 * might as well be placed directly into the zero queue.
	 */
	if (vm_page_sleep(p, "pmaprl", NULL))
		return 0;

	vm_page_busy(p);

	/*
	 * Remove the page table page from the processes address space.
	 */
	pde[p->pindex] = 0;
	pmap->pm_stats.resident_count--;

#if !defined(MAX_PERF)
	if (p->hold_count)  {
		panic("pmap_release: freeing held page table page: index=%d, hc=%d",
			p->pindex, p->hold_count);
	}
#endif
	/*
	 * Page directory pages need to have the kernel
	 * stuff cleared, so they can go into the zero queue also.
	 */
	if (p->pindex == PTDPTDI) {
		bzero(pde + KPTDI, nkpt * PTESIZE);
#ifdef SMP
		setpte(&pde[MPPTDI], 0);
#endif
		setpte(&pde[APTDPTDI], 0);
		pmap_kremove((vm_offset_t) pmap->pm_pdir);
	}

	if (pmap->pm_ptphint == p)
		pmap->pm_ptphint = NULL;

	p->wire_count--;
	cnt.v_wire_count--;
	vm_page_free_zero(p);
	return 1;
}

/*
 * this routine is called if the page table page is not
 * mapped correctly.
 */
static vm_page_t
_pmap_allocpte(pmap, ptepindex)
	pmap_t	pmap;
	unsigned ptepindex;
{
	vm_offset_t pteva, ptepa, ptebva;
	vm_page_t m;

	/*
	 * Find or fabricate a new pagetable page
	 */
	m = vm_page_grab(pmap->pm_pteobj, ptepindex,
			VM_ALLOC_ZERO | VM_ALLOC_RETRY);

	if (m->queue != PQ_NONE) {
		int s = splvm();
		vm_page_unqueue(m, TRUE);
		splx(s);
	}

	/*
	 * Increment the hold count for the page table page
	 * (denoting a new mapping.)
	 */
	if (m->wire_count == 0)
		cnt.v_wire_count++;
	m->wire_count++;
	m->hold_count++;

	/*
	 * Map the pagetable page into the process address space, if
	 * it isn't already there.
	 */

	pmap->pm_stats.resident_count++;

	ptepa = VM_PAGE_TO_PHYS(m);
	ptebva = UPT_MIN_ADDRESS + i386_ptob(ptepindex);
	setpte(&pmap->pm_pdir[ptepindex],
		(ptepa | PG_U | PG_RW | PG_V | PG_A | PG_M));
#if defined(SMP)
	invlpgcpu(ptebva, pmap->pm_pactive);
#else
	invlpg(ptebva);
#endif

	/*
	 * Set the page table hint
	 */
	pmap->pm_ptphint = m;

	/*
	 * Try to use the new mapping, but if we cannot, then
	 * do it with the routine that maps the page explicitly.
	 */
	if ((m->flags & PG_ZERO) == 0) {
		if ((((unsigned)pmap->pm_pdir[PTDPTDI]) & PG_FRAME) ==
			(((unsigned) PTDpde) & PG_FRAME)) {
			pteva = UPT_MIN_ADDRESS + i386_ptob(ptepindex);
			_pmap_zero_page(pteva);
		} else {
			pmap_zero_page(ptepa);
		}
	}

	m->valid = VM_PAGE_BITS_ALL;
	m->flags &= ~PG_ZERO;
	m->flags |= PG_MAPPED;

	return m;
}

static vm_page_t
pmap_allocpte(pmap, va)
	pmap_t	pmap;
	vm_offset_t va;
{
	unsigned ptepindex;
	vm_offset_t ptepa;
	vm_page_t m;

	/*
	 * Calculate pagetable page index
	 */
	ptepindex = va >> PDRSHIFT;

retry:
	/*
	 * Get the page directory entry
	 */
	ptepa = (vm_offset_t) pmap->pm_pdir[ptepindex];

	/*
	 * This supports switching from a 4MB page to a
	 * normal 4K page.
	 */
	if (ptepa & PG_PS) {
		setpte(&pmap->pm_pdir[ptepindex], 0);
		ptepa = 0;
		invltlb();
	}

	/*
	 * If the page table page is mapped, we just increment the
	 * hold count, and activate it.
	 */
	if (ptepa) {
		/*
		 * In order to get the page table page, try the
		 * hint first.
		 */
		m = pmap_getpte(pmap, ptepindex);
		if (!m)
			goto retry;

		m->hold_count++;

#if defined(DIAGNOSTIC)
		if (VM_PAGE_TO_PHYS(m) != (ptepa & PG_FRAME)) {
			panic("pmap_allocpte: physical pte mismatch");
		}
		if (m->wire_count == 0) {
			panic("pmap_allocpte: pte wire count problem");
		}
#endif
		return m;
	}
	/*
	 * Here if the pte page isn't mapped, or if it has been deallocated.
	 */
	return _pmap_allocpte(pmap, ptepindex);
}


/***************************************************
* Pmap allocation/deallocation routines.
 ***************************************************/

/*
 * Release any resources held by the given physical map.
 * Called when a pmap initialized by pmap_pinit is being released.
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_release(pmap)
	register struct pmap *pmap;
{
	vm_page_t p,n,ptdpg;
	vm_object_t object = pmap->pm_pteobj;
	int curgeneration;

#if defined(PMAP_DIAGNOSTIC)
	if (object->ref_count != 1)
		panic("pmap_release: pteobj reference count != 1");
#endif
	
	ptdpg = NULL;
retry:
	curgeneration = object->generation;
	for (p = TAILQ_FIRST(&object->memq); p != NULL; p = n) {
		n = TAILQ_NEXT(p, listq);
		if (p->pindex == PTDPTDI) {
			ptdpg = p;
			continue;
		}
		while (!pmap_release_free_page(pmap, p)) {
			if (object->generation != curgeneration)
				goto retry;
		}
	}

	if (ptdpg && !pmap_release_free_page(pmap, ptdpg))
		goto retry;
}

/*
 * grow the number of kernel page table entries, if needed
 */
void
pmap_growkernel(vm_offset_t addr)
{
	struct proc *p;
	struct pmap *pmap;
	int s;
	vm_offset_t ptppaddr;
	vm_page_t nkpg;
#ifdef SMP
	int i;
#endif
	unsigned int newpdir;

	s = splhigh();
	if (kernel_vm_end == 0) {
		kernel_vm_end = KERNBASE;
		nkpt = 0;
		while (pdir_pde(PTD, kernel_vm_end)) {
			kernel_vm_end = (kernel_vm_end + PAGE_SIZE * NPTEPG) & ~(PAGE_SIZE * NPTEPG - 1);
			nkpt++;
		}
	}
	addr = (addr + PAGE_SIZE * NPTEPG) & ~(PAGE_SIZE * NPTEPG - 1);
	while (kernel_vm_end < addr) {
		if (pdir_pde(PTD, kernel_vm_end)) {
			kernel_vm_end = (kernel_vm_end + PAGE_SIZE * NPTEPG) & ~(PAGE_SIZE * NPTEPG - 1);
			continue;
		}
		if (kptobj == NULL) {
			kptobj = vm_object_allocate(OBJT_DEFAULT, NKPDE);
		}

		/*
		 * This index is bogus, but out of the way
		 */
		nkpg = vm_page_alloc(kptobj, nkpt, VM_ALLOC_SYSTEM);
#if !defined(MAX_PERF)
		if (!nkpg)
			panic("pmap_growkernel: no memory to grow kernel");
#endif

		nkpt++;
		nkpg->flags &= ~PG_ZERO;

		vm_page_wire(nkpg);
		ptppaddr = VM_PAGE_TO_PHYS(nkpg);
		pmap_zero_page(ptppaddr);
		newpdir = ptppaddr | PG_V | PG_RW | PG_A | PG_M;
		setpte(&pdir_pde(PTD, kernel_vm_end), newpdir);

#ifdef SMP
		for (i = 0; i < mp_ncpus; i++) {
			if (IdlePTDS[i])
				setpte(&pdir_pde(IdlePTDS[i], kernel_vm_end), newpdir);
		}
#endif

		for (p = allproc.lh_first; p != 0; p = p->p_list.le_next) {
			if (p->p_vmspace) {
				pmap = &p->p_vmspace->vm_pmap;
				setpte(pmap_pde(pmap, kernel_vm_end), newpdir);
			}
		}
		setpte(pmap_pde(kernel_pmap, kernel_vm_end), newpdir);
		kernel_vm_end = (kernel_vm_end + PAGE_SIZE * NPTEPG) & ~(PAGE_SIZE * NPTEPG - 1);
	}
	splx(s);
}

/*
 *	Retire the given physical map from service.
 *	Should only be called if the map contains
 *	no valid mappings.
 */
void
pmap_destroy(pmap)
	register pmap_t pmap;
{
	int count;

	if (pmap == NULL)
		return;

	count = --pmap->pm_count;
	if (count == 0) {
		pmap_release(pmap);
#if !defined(MAX_PERF)
		panic("destroying a pmap is not yet implemented");
#endif
	}
}

/*
 *	Add a reference to the specified pmap.
 */
void
pmap_reference(pmap)
	pmap_t pmap;
{
	if (pmap != NULL) {
		pmap->pm_count++;
	}
}

/***************************************************
* page management routines.
 ***************************************************/

/*
 * free the pv_entry back to the free list
 */
static PMAP_INLINE void
free_pv_entry(pv)
	pv_entry_t pv;
{
	pv_entry_count--;
	zfreei(pvzone, pv);
}

/*
 * get a new pv_entry, allocating a block from the system
 * when needed.
 * the memory allocation is performed bypassing the malloc code
 * because of the possibility of allocations at interrupt time.
 */
static pv_entry_t
get_pv_entry(void)
{
	pv_entry_count++;
	if (pv_entry_high_water &&
		(pv_entry_count > pv_entry_high_water) &&
		(pmap_pagedaemon_waken == 0)) {
		pmap_pagedaemon_waken = 1;
		wakeup (&vm_pages_needed);
	}
	return zalloci(pvzone);
}

/*
 * This routine is very drastic, but can save the system
 * in a pinch.
 */
void
pmap_collect() {
	pv_table_t *ppv;
	int i;
	vm_offset_t pa;
	vm_page_t m;
	static int warningdone=0;

	if (pmap_pagedaemon_waken == 0)
		return;

	if (warningdone < 5) {
		printf("pmap_collect: collecting pv entries -- suggest increasing PMAP_SHPGPERPROC\n");
		warningdone++;
	}

	for(i = 0; i < pv_npg; i++) {
		ppv = &pv_table[i];
		m = ppv->pv_vm_page;
		if (m->wire_count || m->hold_count || m->busy ||
			(m->flags & PG_BUSY))
			continue;
		m->flags |= PG_BUSY;
		pmap_remove_all(m);
		vm_page_wakeup(m);
	}
	pmap_pagedaemon_waken = 0;
}
	

/*
 * If it is the first entry on the list, it is actually
 * in the header and we must copy the following entry up
 * to the header.  Otherwise we must search the list for
 * the entry.  In either case we free the now unused entry.
 */

static int
pmap_remove_entry(pmap, oldpte, va, mpte)
	struct pmap *pmap;
	unsigned oldpte;
	vm_offset_t va;
	vm_page_t mpte;
{
	vm_page_t m;
	pv_entry_t pv;
	pv_table_t *ppv;
	int rtval;
	int s;

	ppv = pa_to_pvh((oldpte & PG_FRAME));

	s = splvm();
	if (ppv->pv_list_count < pmap->pm_stats.resident_count) {
		for (pv = TAILQ_FIRST(&ppv->pv_list);
			pv;
			pv = TAILQ_NEXT(pv, pv_list)) {
			if (pmap == pv->pv_pmap && va == pv->pv_va) 
				break;
		}
	} else {
		for (pv = TAILQ_FIRST(&pmap->pm_pvlist);
			pv;
			pv = TAILQ_NEXT(pv, pv_plist)) {
			if (va == pv->pv_va) 
				break;
		}
	}

	if (pv == NULL) {
		splx(s);
		return 0;
	}

	m = ppv->pv_vm_page;
	if (mpte && (mpte->hold_count == 0 || mpte->wire_count == 0))
		panic("pmap_remove_entry: paradoxical hold/wire count: %d/%d (0x%x,%d)\n",
			mpte->hold_count, mpte->wire_count, va, mpte->pindex);
	rtval = pmap_unuse_pt(pmap, va, mpte);

	TAILQ_REMOVE(&ppv->pv_list, pv, pv_list);
	ppv->pv_list_count--;

	if ((pv->pv_flags & PV_TRACKMOD) && (oldpte & PG_M))
		m->dirty = VM_PAGE_BITS_ALL;
	if (oldpte & PG_A)
		m->flags |= PG_REFERENCED;

	if (TAILQ_FIRST(&ppv->pv_list) == NULL)
		m->flags &= ~(PG_MAPPED | PG_WRITEABLE);

	TAILQ_REMOVE(&pmap->pm_pvlist, pv, pv_plist);
	free_pv_entry(pv);
	splx(s);
	return rtval;
}

/*
 * Create a pv entry for page at pa for
 * (pmap, va).
 */
static void
pmap_insert_entry(pmap, va, mpte, m)
	pmap_t pmap;
	vm_offset_t va;
	vm_page_t mpte;
	vm_page_t m;
{

	int s;
	pv_entry_t pv;
	pv_table_t *ppv;

	s = splvm();
	pv = get_pv_entry();
	pv->pv_va = va;
	pv->pv_pmap = pmap;
	pv->pv_ptem = mpte;
	pv->pv_flags = 0;
	if (pmap_track_modified(va))
		pv->pv_flags |= PV_TRACKMOD;

	ppv = m->pventry;

	TAILQ_INSERT_TAIL(&pmap->pm_pvlist, pv, pv_plist);
	TAILQ_INSERT_TAIL(&ppv->pv_list, pv, pv_list);
	ppv->pv_list_count++;

	splx(s);
}

/*
 * pmap_remove_pte: do the things to unmap a page in a process
 */
static int
pmap_remove_pte(pmap, ptq, va, mpte)
	struct pmap *pmap;
	unsigned *ptq;
	vm_offset_t va;
	vm_page_t mpte;
{
	unsigned oldpte;
	int rtval;

	oldpte = setpte(ptq, 0);
	/*
	 * Machines that don't support invlpg, also don't support
	 * PG_G.
	 */
	if (oldpte & PG_G)
		invlpg(va);
	if (oldpte & PG_W)
		pmap->pm_stats.wired_count -= 1;


	if (oldpte & PG_MANAGED) {
		rtval = pmap_remove_entry(pmap, oldpte, va, mpte);
	} else {
		rtval = pmap_unuse_pt(pmap, va, mpte);
	}

	pmap->pm_stats.resident_count -= 1;

	return rtval;
}

/*
 * Remove a single page from a process address space
 */
static void
pmap_remove_page(pmap, va)
	struct pmap *pmap;
	register vm_offset_t va;
{
	register unsigned *ptq;
	/*
	 * get a local va for mappings for this pmap.
	 */
	ptq = pmap_pte(pmap, va);
	if (ptq && *ptq) {
		(void) pmap_remove_pte(pmap, ptq, va, NULL);
#if defined(SMP)
		invlpgcpu(va, pmap->pm_pactive);
#else
		invlpg(va);
#endif
	}
	return;
}

/*
 *	Remove the given range of addresses from the specified map.
 *
 *	It is assumed that the start and end are properly
 *	rounded to the page size.
 */
void
pmap_remove(pmap, sva, eva)
	struct pmap *pmap;
	register vm_offset_t sva;
	register vm_offset_t eva;
{
	register unsigned *ptbase;
	unsigned pdirindex;
	vm_offset_t pdnxt;
	vm_offset_t ptpaddr;
	vm_offset_t sindex, eindex;
	vm_page_t mpte;
	int anyvalid;

	if (pmap == NULL)
		return;

	if (pmap->pm_stats.resident_count == 0)
		return;

	/*
	 * special handling of removing one page.  a very
	 * common operation and easy to short circuit some
	 * code.
	 */
	if (((sva + PAGE_SIZE) == eva) && 
		(((unsigned) pmap->pm_pdir[(sva >> PDRSHIFT)] & PG_PS) == 0)) {
		pmap_remove_page(pmap, sva);
		return;
	}

	anyvalid = 0;

	/*
	 * Get a local virtual address for the mappings that are being
	 * worked with.
	 */
	ptbase = get_ptbase(pmap);

	sindex = i386_btop(sva);
	eindex = i386_btop(eva);

	mpte = NULL;
	for (; sindex < eindex; sindex = pdnxt) {

		/*
		 * Calculate index for next page table.
		 */
		pdirindex = sindex / NPDEPG;
		pdnxt = ((sindex + NPTEPG) & ~(NPTEPG - 1));

		if (pmap != kernel_pmap) {
			if (pmap->pm_stats.resident_count == 0)
				break;

			if ((mpte == NULL) || (mpte->pindex != pdirindex)) {
				page_wakeupq(&mpte);
				mpte = pmap_getpte(pmap, pdirindex);
			}
		}

		if (((ptpaddr = (unsigned) pmap->pm_pdir[pdirindex]) & PG_PS) != 0) {
			vm_offset_t va;
			setpte(&pmap->pm_pdir[pdirindex], 0);
			va = UPT_MIN_ADDRESS + i386_ptob(sindex);
#if defined(SMP)
			invlpgcpu(va, pmap->pm_pactive);
#else
			invlpg(va);
#endif
			pmap->pm_stats.resident_count -= NBPDR / PAGE_SIZE;
			anyvalid++;
			continue;
		}

		/*
		 * Weed out invalid mappings. Note: we assume that the page
		 * directory table is always allocated, and in kernel virtual.
		 */
		if (ptpaddr == 0)
			continue;

		/*
		 * Limit our scan to either the end of the va represented
		 * by the current page table page, or to the end of the
		 * range being removed.
		 */
		if (pdnxt > eindex) {
			pdnxt = eindex;
		}

		for ( ;sindex != pdnxt; sindex++) {
			vm_offset_t va;
			if (ptbase[sindex] == 0) {
				continue;
			}
			va = i386_ptob(sindex);
			
			anyvalid++;
			if (pmap_remove_pte(pmap,
				ptbase + sindex, va, mpte)) {
				mpte = NULL;
				break;
			}
		}
	}
	page_wakeupq(&mpte);

#if defined(SMP)
	if (anyvalid)
		invltlbcpu(pmap->pm_pactive);
#else
	if (anyvalid)
		invltlb();
#endif
}

/*
 *	Routine:	pmap_remove_all
 *	Function:
 *		Removes this physical page from
 *		all physical maps in which it resides.
 *		Reflects back modify bits to the pager.
 *
 *	Notes:
 *		Original versions of this routine were very
 *		inefficient because they iteratively called
 *		pmap_remove (slow...)
 */

static void
pmap_remove_all(m)
	vm_page_t m;
{
	register pv_entry_t pv;
	vm_offset_t pa;
	pv_table_t *ppv;
	register unsigned *pte, tpte;
	vm_page_t mpte;
	int update_needed;
	int s;

	update_needed = 0;
#if defined(PMAP_DIAGNOSTIC)
	/*
	 * XXX this makes pmap_page_protect(NONE) illegal for non-managed
	 * pages!
	 */
	if ((m->flags & PG_FICTITIOUS) || (m->pventry == NULL)) {
		panic("pmap_page_protect: illegal for unmanaged page, va: 0x%x", VM_PAGE_TO_PHYS(m));
	}
#endif

	s = splvm();
	ppv = m->pventry;
retry:
	while ((pv = TAILQ_FIRST(&ppv->pv_list)) != NULL) {
		pmap_t pmap;
		mpte = pv->pv_ptem;

		if (mpte) {
			if (vm_page_sleep(mpte,"pmrma",NULL))
				goto retry;
			vm_page_busy(mpte);
		}

		pmap = pv->pv_pmap;
		pte = pmap_pte(pmap, pv->pv_va);

		pmap->pm_stats.resident_count--;

		tpte = setpte(pte, 0);

		if (tpte & PG_W)
			pmap->pm_stats.wired_count--;

		if (tpte & PG_A)
			m->flags |= PG_REFERENCED;

		/*
		 * Update the vm_page_t clean and reference bits.
		 */
		if (tpte & PG_M) {
			if (pv->pv_flags & PV_TRACKMOD)
				m->dirty = VM_PAGE_BITS_ALL;
		}
		update_needed |= pmap->pm_pactive;

		TAILQ_REMOVE(&pmap->pm_pvlist, pv, pv_plist);
		TAILQ_REMOVE(&ppv->pv_list, pv, pv_list);
		ppv->pv_list_count--;
		if (!pmap_unuse_pt(pmap, pv->pv_va, mpte))
			page_wakeupq(&mpte);
		free_pv_entry(pv);
	}

	m->flags &= ~(PG_MAPPED | PG_WRITEABLE);

#if defined(SMP)
	if (update_needed)
		invltlbcpu(update_needed);
#else
	if (update_needed)
		invltlb();
#endif

	splx(s);
	return;
}

/*
 *	Set the physical protection on the
 *	specified range of this map as requested.
 */
void
pmap_protect(pmap_t pmap, vm_offset_t sva, vm_offset_t eva, vm_prot_t prot)
{
	unsigned *ptbase;
	vm_offset_t pdnxt, ptpaddr;
	vm_pindex_t sindex, eindex;
	int update_needed;
	pv_table_t *ppv;
	pv_entry_t pv;


	if (pmap == NULL)
		return;

	if ((prot & VM_PROT_READ) == VM_PROT_NONE) {
		pmap_remove(pmap, sva, eva);
		return;
	}

	if (prot & VM_PROT_WRITE)
		return;

	update_needed = 0;

	ptbase = get_ptbase(pmap);

	sindex = i386_btop(sva);
	eindex = i386_btop(eva);

	for (; sindex < eindex; sindex = pdnxt) {

		unsigned pdirindex;

		pdnxt = ((sindex + NPTEPG) & ~(NPTEPG - 1));

		pdirindex = sindex / NPDEPG;
		if (((ptpaddr = (unsigned) pmap->pm_pdir[pdirindex]) & PG_PS) != 0) {
			(unsigned) pmap->pm_pdir[pdirindex] &= ~(PG_M|PG_RW);
			pmap->pm_stats.resident_count -= NBPDR / PAGE_SIZE;
			update_needed |= pmap->pm_pactive;
			continue;
		}

		/*
		 * Weed out invalid mappings. Note: we assume that the page
		 * directory table is always allocated, and in kernel virtual.
		 */
		if (ptpaddr == 0)
			continue;

		if (pdnxt > eindex) {
			pdnxt = eindex;
		}

		for (; sindex != pdnxt; sindex++) {

			unsigned pbits;
			vm_page_t m;
			pv_table_t *ppv;

			pbits = ptbase[sindex];

			if (pbits & PG_MANAGED) {
				if (pbits & (PG_A|PG_M)) {
					ppv = pa_to_pvh(pbits);
					m = ppv->pv_vm_page;
					if ((pbits & PG_M) == 0) {
						m->flags |= PG_REFERENCED;
					} else {
						if (pmap_track_modified(i386_ptob(sindex))) {
							m->dirty = VM_PAGE_BITS_ALL;
						}
						if (pbits & PG_A) {
							m->flags |= PG_REFERENCED;
						}
					}
					pbits &= ~(PG_A|PG_M);
					pv = TAILQ_FIRST(&ppv->pv_list);
					if (pv && (TAILQ_NEXT(pv, pv_list) == NULL) &&
						(pv->pv_pmap == pmap)) {
						vm_offset_t tva;
						tva = i386_ptob(sindex);
						if (pv->pv_va == tva) {
							m->flags &= ~PG_WRITEABLE;
						}
					}
				}
			}

			pbits &= ~PG_RW;

			if (pbits != ptbase[sindex]) {
				setpte(&ptbase[sindex], pbits);
				update_needed |= pmap->pm_pactive;
			}
		}
	}
#if defined(SMP)
	if (update_needed)
		invltlbcpu(update_needed);
#else
	if (update_needed)
		invltlb();
#endif
}

/*
 *	Insert the given physical page (p) at
 *	the specified virtual address (v) in the
 *	target physical map with the protection requested.
 *
 *	If specified, the page will be wired down, meaning
 *	that the related pte can not be reclaimed.
 *
 *	NB:  This is the only routine which MAY NOT lazy-evaluate
 *	or lose information.  That is, this routine must actually
 *	insert this page into the given map NOW.
 */
void
pmap_enter(pmap_t pmap, vm_offset_t va, vm_page_t m, vm_prot_t prot,
	   boolean_t wired)
{
	vm_offset_t pa;
	register unsigned *pte;
	vm_offset_t opa;
	vm_offset_t origpte, newpte;
	vm_page_t mpte;
	pv_table_t *ppv;

	if (pmap == NULL)
		return;

	va &= PG_FRAME;
#ifdef PMAP_DIAGNOSTIC
	if (va > VM_MAX_KERNEL_ADDRESS)
		panic("pmap_enter: toobig");
	if ((va >= UPT_MIN_ADDRESS) && (va < UPT_MAX_ADDRESS))
		panic("pmap_enter: invalid to pmap_enter page table pages (va: 0x%x)", va);
#endif

	/*
	 * In the case that a page table page is not
	 * resident, we are creating it here.
	 */
	if (va < UPT_MIN_ADDRESS) {
		mpte = pmap_allocpte(pmap, va);
		if (curproc && curproc->p_vmspace &&
			(pmap != &curproc->p_vmspace->vm_pmap)) {
			pte = pmap_pte(pmap, va);
		} else {
			pte = (unsigned *)vtopte(va);
		}
	} else {
		mpte = NULL;
		pte = (unsigned *)vtopte(va);
	}

#if defined(DIAGNOSTIC) && !defined(MAX_PERF)
	/*
	 * Page Directory table entry not valid, we need a new PT page
	 */
	if (pte == NULL) {
		panic("pmap_enter: invalid page directory, pdir=%p, va=0x%x\n",
			(void *)pmap->pm_pdir[PTDPTDI], va);
	}
#endif

	pa = VM_PAGE_TO_PHYS(m) & PG_FRAME;
	origpte = *(vm_offset_t *)pte;
	opa = origpte & PG_FRAME;

#if defined(DIAGNOSTIC) && !defined(MAX_PERF)
	if (origpte & PG_PS) {
		printf("VA: 0x%x, PA: 0x%x, Orig: 0x%x, PROT: 0x%x, Wired: %d, pseflag: %x\n",
			va, pa, origpte, prot, wired, pseflag);
		if (mpte) {
			printf("Pte hold count: %d, pindex: %d\n", mpte->hold_count,
				mpte->pindex);
		}
		panic("pmap_enter: attempted pmap_enter on 4MB page");
	}
#endif

	/*
	 * Mapping has not changed, must be protection or wiring change.
	 */
	if (origpte && (opa == pa)) {
		/*
		 * Wiring change, just update stats. We don't worry about
		 * wiring PT pages as they remain resident as long as there
		 * are valid mappings in them. Hence, if a user page is wired,
		 * the PT page will be also.
		 */
		if (wired && ((origpte & PG_W) == 0))
			pmap->pm_stats.wired_count++;
		else if (!wired && (origpte & PG_W))
			pmap->pm_stats.wired_count--;

#if defined(PMAP_DIAGNOSTIC)
		if (pmap_nw_modified((pt_entry_t) origpte)) {
			printf(
	"pmap_enter: modified page not writable: va: 0x%x, pte: 0x%x\n",
			    va, origpte);
		}
#endif

		/*
		 * Remove extra pte reference
		 */
		if (mpte)
			mpte->hold_count--;

		/*
		 * Turning on write access
		 */
		if ((prot & VM_PROT_WRITE) && ((origpte & (PG_RW|PG_V)) == PG_V))
			goto validate;

		/*
		 * We might be turning off write access to the page,
		 * so we go ahead and sense modify status.
		 */
		if ((origpte & (PG_MANAGED|PG_M)) == (PG_MANAGED|PG_M)) {
			if ((m->dirty != VM_PAGE_BITS_ALL) &&
				pmap_track_modified(va)) {
				origpte &= ~PG_M;
				m->dirty = VM_PAGE_BITS_ALL;
			}
		}
		goto validate;
	} 
	/*
	 * Mapping has changed, invalidate old range and fall through to
	 * handle validating new mapping.
	 */
	if (opa) {
		int err;
		err = pmap_remove_pte(pmap, pte, va, mpte);
#if !defined(MAX_PERF)
		if (err)
			panic("pmap_enter: pte vanished, va: 0x%x", va);
#endif
	}

	/*
	 * Enter on the PV list if part of our managed memory Note that we
	 * raise IPL while manipulating pv_table since pmap_enter can be
	 * called at interrupt time.
	 */
	if ((m->flags & PG_FICTITIOUS) == 0)
		pmap_insert_entry(pmap, va, mpte, m);

	/*
	 * Increment counters
	 */
	pmap->pm_stats.resident_count++;
	if (wired)
		pmap->pm_stats.wired_count++;

validate:
	/*
	 * Now validate mapping with desired protection/wiring.
	 */
	if (prot & VM_PROT_WRITE) {
		newpte = (vm_offset_t) (pa | PG_RW | PG_A | PG_V);
		if (prot & VM_PROT_DIRTY) {
			newpte |= PG_M;
			m->dirty = VM_PAGE_BITS_ALL;
		}
		m->flags |= PG_MAPPED | PG_WRITEABLE;
	} else {
		newpte = (vm_offset_t) (pa | PG_A | PG_V);
		m->flags |= PG_MAPPED;
	}

	if ((m->flags & PG_FICTITIOUS) == 0)
		newpte |= PG_MANAGED;
	if (wired)
		newpte |= PG_W;
	if (va < UPT_MIN_ADDRESS)
		newpte |= PG_U;
#if !defined(SMP)
	if (pmap == kernel_pmap)
		newpte |= pgeflag;
#endif

	/*
	 * if the mapping or permission bits are different, we need
	 * to update the pte.
	 */
	if ((origpte & ~(PG_M|PG_A)) != (newpte & ~(PG_M|PG_A))) {
		setpte(pte, newpte);
		if (origpte)
#if defined(SMP)
				invlpgcpu(va, pmap->pm_pactive);
#else
				invlpg(va);
#endif
	}
	page_wakeupq(&mpte);
}

/*
 * this code makes some *MAJOR* assumptions:
 * 1. Current pmap & pmap exists.
 * 2. Not wired.
 * 3. Read access.
 * 4. No page table pages.
 * 5. Tlbflush is deferred to calling procedure.
 * 6. Page IS managed.
 * but is *MUCH* faster than pmap_enter...
 */

static vm_page_t
pmap_enter_quick(pmap, va, m, mpte)
	pmap_t pmap;
	vm_offset_t va;
	vm_page_t m;
	vm_page_t mpte;
{
	unsigned *pte;
	unsigned ptepindex;
	vm_offset_t pa;

	/*
	 * This call to vtopte makes the assumption that we are
	 * entering the page into the current pmap.  In order to support
	 * quick entry into any pmap, one would likely use pmap_pte_quick.
	 * But that isn't as quick as vtopte.
	 */
	pte = (unsigned *)vtopte(va);

	/*
	 * In the case that a page table page is not
	 * resident, we are creating it here.
	 */
	if (va < UPT_MIN_ADDRESS) {
		/*
		 * Calculate pagetable page index
		 */
		ptepindex = va >> PDRSHIFT;

		if (mpte && (mpte->pindex == ptepindex) && mpte->wire_count) {
			mpte->hold_count++;
		} else {
			page_wakeupq(&mpte);
			mpte = pmap_allocpte(pmap, va);
		}
	} else {
		page_wakeupq(&mpte);
	}

	if (*pte) {
		if (pmap_unwire_pte_hold(pmap, mpte)) {
			return 0;
		}
		return mpte;
	}

	/*
	 * Enter on the PV list if part of our managed memory Note that we
	 * raise IPL while manipulating pv_table since pmap_enter can be
	 * called at interrupt time.
	 */
	pmap_insert_entry(pmap, va, mpte, m);

	/*
	 * Increment counters
	 */
	pmap->pm_stats.resident_count++;

	pa = VM_PAGE_TO_PHYS(m);
	m->flags |= PG_MAPPED;

	/*
	 * Now validate mapping with RO protection
	 */
	if ((m->flags & PG_FICTITIOUS) == 0)
		setpte(pte, pa | PG_V | PG_U | PG_MANAGED);
	else
		setpte(pte, pa | PG_V | PG_U);

	return mpte;
}

#define MAX_INIT_PT (96)
/*
 * pmap_object_init_pt preloads the ptes for a given object
 * into the specified pmap.  This eliminates the blast of soft
 * faults on process startup and immediately after an mmap.
 */
void
pmap_object_init_pt(pmap, addr, object, pindex, size, limit)
	pmap_t pmap;
	vm_offset_t addr;
	vm_object_t object;
	vm_pindex_t pindex;
	vm_size_t size;
	int limit;
{
	vm_offset_t tmpidx;
	int psize;
	vm_page_t p, mpte;
	int objpgs;

	if (!pmap)
		return;

	/*
	 * This code maps large physical mmap regions into the
	 * processor address space.  Note that some shortcuts
	 * are taken, but the code works.
	 */
	if (pseflag &&
		(object->type == OBJT_DEVICE) &&
		((addr & (NBPDR - 1)) == 0) &&
		((size & (NBPDR - 1)) == 0) ) {
		int i;
		int s;
		vm_page_t m[1];
		unsigned int ptepindex;
		int npdes;
		vm_offset_t ptepa;

		if (pmap->pm_pdir[ptepindex = (addr >> PDRSHIFT)])
			return;

retry:
		p = vm_page_lookup(object, pindex);
		if (p && vm_page_sleep(p, "init4p", NULL))
			goto retry;

		if (p == NULL) {
			p = vm_page_alloc(object, pindex, VM_ALLOC_NORMAL);
			if (p == NULL)
				return;
			m[0] = p;

			if (vm_pager_get_pages(object, m, 1, 0) != VM_PAGER_OK) {
				vm_page_free(p);
				return;
			}

			p = vm_page_lookup(object, pindex);
			vm_page_wakeup(p);
		}

		ptepa = (vm_offset_t) VM_PAGE_TO_PHYS(p);
		if (ptepa & (NBPDR - 1)) {
			return;
		}

		p->valid = VM_PAGE_BITS_ALL;
		p->flags &= ~PG_ZERO;

		pmap->pm_stats.resident_count += size >> PAGE_SHIFT;
		npdes = size >> PDRSHIFT;
		for(i=0;i<npdes;i++) {
			setpte(&pmap->pm_pdir[ptepindex], ptepa | PG_U | PG_RW |
					PG_V | PG_M | PG_A | PG_PS);
			ptepa += NBPDR;
			ptepindex += 1;
		}
		p->flags |= PG_MAPPED;
		invltlb();
		return;
	}

	psize = i386_btop(size);

	if ((object->type != OBJT_VNODE) ||
		(limit && (psize > MAX_INIT_PT) &&
			(object->resident_page_count > MAX_INIT_PT))) {
		return;
	}

	if (psize + pindex > object->size)
		psize = object->size - pindex;

	mpte = NULL;
	/*
	 * if we are processing a major portion of the object, then scan the
	 * entire thing.
	 */
	if (psize > (object->size >> 2)) {
		objpgs = psize;

		for (p = TAILQ_FIRST(&object->memq);
		    ((objpgs > 0) && (p != NULL));
		    p = TAILQ_NEXT(p, listq)) {

			tmpidx = p->pindex;
			if (tmpidx < pindex) {
				continue;
			}
			tmpidx -= pindex;
			if (tmpidx >= psize) {
				continue;
			}
			if (((p->valid & VM_PAGE_BITS_ALL) == VM_PAGE_BITS_ALL) &&
			    (p->flags & (PG_BUSY | PG_FICTITIOUS)) == 0) {
				if ((p->queue - p->pc) == PQ_CACHE)
					vm_page_deactivate(p);
				vm_page_busy(p);
				mpte = pmap_enter_quick(pmap, 
					addr + i386_ptob(tmpidx), p, mpte);
				vm_page_wakeup(p);
			}
			objpgs -= 1;
		}
	} else {
		/*
		 * else lookup the pages one-by-one.
		 */
		for (tmpidx = 0; tmpidx < psize; tmpidx += 1) {
			p = vm_page_lookup(object, tmpidx + pindex);
			if (p &&
			    ((p->valid & VM_PAGE_BITS_ALL) == VM_PAGE_BITS_ALL) &&
			     (p->flags & (PG_BUSY | PG_FICTITIOUS)) == 0) {
				if ((p->queue - p->pc) == PQ_CACHE)
					vm_page_deactivate(p);
				vm_page_busy(p);
				mpte = pmap_enter_quick(pmap, 
					addr + i386_ptob(tmpidx), p, mpte);
				vm_page_wakeup(p);
			}
		}
	}
	page_wakeupq(&mpte);
	return;
}

/*
 * pmap_prefault provides a quick way of clustering
 * pagefaults into a processes address space.  It is a "cousin"
 * of pmap_object_init_pt, except it runs at page fault time instead
 * of mmap time.
 */
#define PFBAK 4
#define PFFOR 4
#define PAGEORDER_SIZE (PFBAK+PFFOR)

static int pmap_prefault_pageorder[] = {
	-PAGE_SIZE, PAGE_SIZE,
	-2 * PAGE_SIZE, 2 * PAGE_SIZE,
	-3 * PAGE_SIZE, 3 * PAGE_SIZE
	-4 * PAGE_SIZE, 4 * PAGE_SIZE
};

void
pmap_prefault(pmap, addra, entry)
	pmap_t pmap;
	vm_offset_t addra;
	vm_map_entry_t entry;
{
	int i;
	vm_offset_t starta;
	vm_offset_t addr;
	vm_pindex_t pindex;
	vm_page_t m, mpte;
	vm_object_t object;

	if (!curproc || (pmap != &curproc->p_vmspace->vm_pmap))
		return;

	object = entry->object.vm_object;

	starta = addra - PFBAK * PAGE_SIZE;
	if (starta < entry->start) {
		starta = entry->start;
	} else if (starta > addra) {
		starta = 0;
	}

	mpte = NULL;
	for (i = 0; i < PAGEORDER_SIZE; i++) {
		vm_object_t lobject;
		unsigned *pte;

		addr = addra + pmap_prefault_pageorder[i];
		if (addr > addra + (PFFOR * PAGE_SIZE))
			addr = 0;

		if (addr < starta || addr >= entry->end)
			continue;

		if ((*pmap_pde(pmap, addr)) == NULL) 
			continue;

		pte = (unsigned *) vtopte(addr);
		if (*pte)
			continue;

		pindex = ((addr - entry->start) + entry->offset) >> PAGE_SHIFT;
		lobject = object;
		for (m = vm_page_lookup(lobject, pindex);
		    (!m && (lobject->type == OBJT_DEFAULT) && (lobject->backing_object));
		    lobject = lobject->backing_object) {
			if (lobject->backing_object_offset & PAGE_MASK)
				break;
			pindex += (lobject->backing_object_offset >> PAGE_SHIFT);
			m = vm_page_lookup(lobject->backing_object, pindex);
		}

		/*
		 * give-up when a page is not in memory
		 */
		if (m == NULL)
			break;

		if (((m->valid & VM_PAGE_BITS_ALL) == VM_PAGE_BITS_ALL) &&
		    (m->flags & (PG_BUSY | PG_FICTITIOUS)) == 0) {

			if ((m->queue - m->pc) == PQ_CACHE) {
				vm_page_deactivate(m);
			}
			vm_page_busy(m);
			mpte = pmap_enter_quick(pmap, addr, m, mpte);
			vm_page_wakeup(m);
		}
	}
	page_wakeupq(&mpte);
}

/*
 *	Routine:	pmap_change_wiring
 *	Function:	Change the wiring attribute for a map/virtual-address
 *			pair.
 *	In/out conditions:
 *			The mapping must already exist in the pmap.
 */
void
pmap_change_wiring(pmap, va, wired)
	register pmap_t pmap;
	vm_offset_t va;
	boolean_t wired;
{
	register unsigned *pte;

	if (pmap == NULL)
		return;

	pte = pmap_pte(pmap, va);
	if (pte == NULL)
		return;

	if (wired && !pmap_pte_w(pte))
		pmap->pm_stats.wired_count++;
	else if (!wired && pmap_pte_w(pte))
		pmap->pm_stats.wired_count--;

	/*
	 * Wiring is not a hardware characteristic so there is no need to
	 * invalidate TLB.
	 */
	pmap_pte_set_w(pte, wired);
}

static __inline void
setaltframe(vm_offset_t frame)
{
	if (frame != (((unsigned) APTDpde) & PG_FRAME)) {
		setpte(&APTDpde, frame | PG_RW | PG_V | PG_A);
#if defined(SMP)
		cpu_invltlb();
#else
		invltlb();
#endif
	}
}

/*
 *	Copy the range specified by src_addr/len
 *	from the source map to the range dst_addr/len
 *	in the destination map.
 *
 *	This routine is only advisory and need not do anything.
 */

void
pmap_copy(dst_pmap, src_pmap, dst_addr, len, src_addr)
	pmap_t dst_pmap, src_pmap;
	vm_offset_t dst_addr;
	vm_size_t len;
	vm_offset_t src_addr;
{
	vm_offset_t addr;
	vm_offset_t end_addr = src_addr + len;
	vm_offset_t pdnxt;
	vm_page_t m;
	int temp;
	unsigned src_frame, dst_frame;

	if (dst_addr != src_addr)
		return;

	src_frame = ((unsigned) src_pmap->pm_pdir[PTDPTDI]) & PG_FRAME;
	if (src_frame != (((unsigned) PTDpde) & PG_FRAME)) {
		return;
	}
	dst_frame = ((unsigned) dst_pmap->pm_pdir[PTDPTDI]) & PG_FRAME;
	setaltframe(dst_frame);

	for(addr = src_addr; addr < end_addr; addr = pdnxt) {
		unsigned *src_pte, *dst_pte, ptetemp, ptepindex;
		vm_page_t dstmpte, srcmpte;
		pv_table_t *ppv;
		vm_offset_t srcptepaddr;

#if !defined(MAX_PERF)
		if (addr >= UPT_MIN_ADDRESS)
			panic("pmap_copy: invalid to pmap_copy page tables\n");
#endif

		/*
		 * Don't let optional prefaulting of pages make us go
		 * way below the low water mark of free pages or way
		 * above high water mark of used pv entries.
		 */
		if (cnt.v_free_count < cnt.v_free_reserved ||
		    pv_entry_count > pv_entry_high_water)
			break;
		
		ptepindex = addr >> PDRSHIFT;
		pdnxt = (ptepindex + 1) << PDRSHIFT;

		srcptepaddr = (vm_offset_t) src_pmap->pm_pdir[ptepindex];
		if ((srcptepaddr & PG_V) == 0)
			continue;
			
		if (srcptepaddr & PG_PS) {
			if (dst_pmap->pm_pdir[ptepindex] == 0) {
				dst_pmap->pm_pdir[ptepindex] = (pd_entry_t) srcptepaddr;
				dst_pmap->pm_stats.resident_count += NBPDR;
			}
			continue;
		}

	retry:
		srcmpte = vm_page_lookup(src_pmap->pm_pteobj, ptepindex);
		if ((srcmpte == NULL) ||
			(srcmpte->hold_count == 0) || (srcmpte->flags & PG_BUSY))
			continue;

		vm_page_busy(srcmpte);

		if (pdnxt > end_addr)
			pdnxt = end_addr;

		src_pte = (unsigned *) vtopte(addr);
		dst_pte = (unsigned *) avtopte(addr);
		while (addr < pdnxt) {

			ptetemp = *src_pte;

			/*
			 * we only virtual copy managed pages
			 */
			if ((ptetemp & (PG_MANAGED|PG_V)) == (PG_MANAGED|PG_V)) {

				ppv = pa_to_pvh(ptetemp);
				m = ppv->pv_vm_page;

				if ((m->flags & (PG_BUSY|PG_FICTITIOUS)) == 0) {

					vm_page_busy(m);
					/*
					 * We have to check after allocpte for the
					 * pte still being around...  allocpte can
					 * block.
					 */
					dstmpte = pmap_allocpte(dst_pmap, addr);
					setaltframe(dst_frame);

					if (*dst_pte == 0) {
						/*
						 * Clear the modified and
						 * accessed (referenced) bits
						 * during the copy.
						 */
						pmap_insert_entry(dst_pmap, addr, dstmpte, m);
						setaltframe(dst_frame);
						setpte(dst_pte, ptetemp & ~(PG_M|PG_A));
						dst_pmap->pm_stats.resident_count++;
						vm_page_wakeup(m);
						vm_page_wakeup(dstmpte);
		 			} else {
						vm_page_wakeup(m);
						if (!pmap_unwire_pte_hold(dst_pmap, dstmpte)) {
							vm_page_wakeup(dstmpte);
						}
					}
					if (dstmpte->hold_count >= srcmpte->hold_count)
						break;
				}
			}
			addr += PAGE_SIZE;
			src_pte++;
			dst_pte++;
		}
		vm_page_wakeup(srcmpte);
	}
}	

/*
 *	Routine:	pmap_kernel
 *	Function:
 *		Returns the physical map handle for the kernel.
 */
pmap_t
pmap_kernel()
{
	return (kernel_pmap);
}

/*
 *	pmap_zero_page zeros the specified (machine independent)
 *	page by mapping the page into virtual memory and using
 *	bzero to clear its contents, one machine dependent page
 *	at a time.
 */
void
pmap_zero_page(phys)
	vm_offset_t phys;
{
#ifdef SMP
#if !defined(MAX_PERF)
	if (*(int *) prv_CMAP3)
		panic("pmap_zero_page: prv_CMAP3 busy");
#endif

	setpte(prv_CMAP3, PG_V | PG_RW | (phys & PG_FRAME) | PG_A | PG_M);
	cpu_invlpg((vm_offset_t)&prv_CPAGE3);
	_pmap_zero_page((vm_offset_t) &prv_CPAGE3);
	setpte(prv_CMAP3, 0);

#else /* !SMP */

#if !defined(MAX_PERF)
	if (*(int *) CMAP2)
		panic("pmap_zero_page: CMAP2 busy");
#endif

	setpte(CMAP2, PG_V | PG_RW | (phys & PG_FRAME) | PG_A | PG_M);
	invlpg((vm_offset_t) CADDR2);
	_pmap_zero_page((vm_offset_t)CADDR2);
	setpte(CMAP2, 0);

#endif
}

#if defined(SMP)
void
pmap_zero_page_alt(phys)
	vm_offset_t phys;
{
	setpte(prv_CMAP4, PG_V | PG_RW | (phys & PG_FRAME) | PG_A | PG_M);
	cpu_invlpg((vm_offset_t)&prv_CPAGE4);
	_pmap_zero_page((vm_offset_t) &prv_CPAGE4);
	setpte(prv_CMAP4, 0);
}
#else
pmap_zero_page_alt(phys)
	vm_offset_t phys;
{
	pmap_zero_page(phys);
}
#endif

/*
 *	pmap_copy_page copies the specified (machine independent)
 *	page by mapping the page into virtual memory and using
 *	bcopy to copy the page, one machine dependent page at a
 *	time.
 */
void
pmap_copy_page(src, dst)
	vm_offset_t src;
	vm_offset_t dst;
{
#ifdef SMP
#if !defined(MAX_PERF)
	if (*(int *) prv_CMAP1)
		panic("pmap_copy_page: prv_CMAP1 busy");
	if (*(int *) prv_CMAP2)
		panic("pmap_copy_page: prv_CMAP2 busy");
#endif

	setpte(prv_CMAP1, PG_V | (src & PG_FRAME) | PG_A);
	cpu_invlpg((vm_offset_t)&prv_CPAGE1);

	setpte(prv_CMAP2, PG_V | PG_RW | (dst & PG_FRAME) | PG_A | PG_M);
	cpu_invlpg((vm_offset_t)&prv_CPAGE2);

	bcopy(&prv_CPAGE1, &prv_CPAGE2, PAGE_SIZE);

	setpte(prv_CMAP1, 0);
	setpte(prv_CMAP2, 0);
#else /* !SMP */
#if !defined(MAX_PERF)
	if (*(int *) CMAP1 || *(int *) CMAP2)
		panic("pmap_copy_page: CMAP busy");
#endif

	if (cpu_class == CPUCLASS_386) {
		setpte(CMAP1, PG_V | (src & PG_FRAME) | PG_A);
		setpte(CMAP2, PG_V | PG_RW | (dst & PG_FRAME) | PG_A | PG_M);
		invltlb();
	} else {
		setpte(CMAP1, PG_V | (src & PG_FRAME) | PG_A);
		invlpg((u_int)CADDR1);
		setpte(CMAP2, PG_V | PG_RW | (dst & PG_FRAME) | PG_A | PG_M);
		invlpg((u_int)CADDR2);
	}

	bcopy(CADDR1, CADDR2, PAGE_SIZE);

	setpte(CMAP1, 0);
	setpte(CMAP2, 0);
#endif
}


/*
 *	Routine:	pmap_pageable
 *	Function:
 *		Make the specified pages (by pmap, offset)
 *		pageable (or not) as requested.
 *
 *		A page which is not pageable may not take
 *		a fault; therefore, its page table entry
 *		must remain valid for the duration.
 *
 *		This routine is merely advisory; pmap_enter
 *		will specify that these pages are to be wired
 *		down (or not) as appropriate.
 */
void
pmap_pageable(pmap, sva, eva, pageable)
	pmap_t pmap;
	vm_offset_t sva, eva;
	boolean_t pageable;
{
}

/*
 * this routine returns true if a physical page resides
 * in the given pmap.
 */
boolean_t
pmap_page_exists(pmap, m)
	pmap_t pmap;
	vm_page_t m;
{
	register pv_entry_t pv;
	pv_table_t *ppv;
	int s;

	if (m->flags & PG_FICTITIOUS)
		return FALSE;

	ppv = m->pventry;
	if (!ppv)
		return FALSE;

	s = splvm();
	/*
	 * Not found, check current mappings returning immediately if found.
	 */
	for (pv = TAILQ_FIRST(&ppv->pv_list);
	     pv;
	     pv = TAILQ_NEXT(pv, pv_list)) {
		if (pv->pv_pmap == pmap) {
			splx(s);
			return TRUE;
		}
	}
	splx(s);
	return (FALSE);
}

#define PMAP_REMOVE_PAGES_CURPROC_ONLY
/*
 * Remove all pages from specified address space
 * this aids process exit speeds.  Also, this code
 * is special cased for current process only, but
 * can have the more generic (and slightly slower)
 * mode enabled.  This is much faster than pmap_remove
 * in the case of running down an entire address space.
 */
void
pmap_remove_pages(pmap, sva, eva)
	pmap_t pmap;
	vm_offset_t sva, eva;
{
	unsigned *pte, tpte;
	pv_table_t *ppv;
	pv_entry_t pv, npv;
	vm_page_t mpte;
	vm_page_t m;
	vm_offset_t va;
	int s;

#if defined(DIAGNOSTIC)
	unsigned pdirindex;

/*
 * Don't mess with kernel_pmap -- shouldn't be called with kernel_pmap
 * anyway.
 */
	if (pmap == kernel_pmap)
		return;
#endif

#ifdef PMAP_REMOVE_PAGES_CURPROC_ONLY
	if (!curproc || (pmap != &curproc->p_vmspace->vm_pmap)) {
		printf("warning: pmap_remove_pages called with non-current pmap\n");
		return;
	}
#endif

	mpte = NULL;
	s = splvm();
retry:
	for(pv = TAILQ_FIRST(&pmap->pm_pvlist); pv; pv = npv) {

		npv = TAILQ_NEXT(pv, pv_plist);
		if (pv->pv_va >= eva || pv->pv_va < sva) {
			continue;
		}


#if defined(DIAGNOSTIC)
		pdirindex = (pv->pv_va >> PDRSHIFT);

		/*
		 * Make sure that we have a pdir entry
		 */
		if (pmap->pm_pdir[pdirindex] == 0) {
			printf("pmap_remove_pages: missing pde (%d), va: 0x%lx\n",
				pdirindex, pv->pv_va);
			continue;
		}
#endif

#if defined(PMAP_REMOVE_PAGES_CURPROC_ONLY)
		pte = (unsigned *)vtopte(pv->pv_va);
#else
		pte = pmap_pte(pmap, pv->pv_va);
#endif

#if defined(DIAGNOSTIC)
		/*
		 * Another way of making sure that there is a pdir
		 */
		if (pte == NULL) {
			printf("pmap_remove_pages: missing pte: va: 0x%x\n", pv->pv_va);
			continue;
		}
#endif

		/*
		 * Switch our view of pte pages
		 */
		if (mpte != pv->pv_ptem) {

			page_wakeupq(&mpte);
			mpte = pv->pv_ptem;

			if (vm_page_sleep(mpte, "pmrvpg", NULL)) {
				mpte = NULL;
				goto retry;
			}

			vm_page_busy(mpte);
		}

#if defined(DIAGNOSTIC)
		/*
		 * Compare our page object index with page directory index
		 */
		if (mpte->pindex != pdirindex) {
			printf("pmap_remove_pages: pindex mismatch: pindex: %d, req: %d\n",
				mpte->pindex, pdirindex);
			continue;
		}

		/*
		 * Should have hold_count per page table entry
		 */
		if (mpte->hold_count == 0) {
			printf("pmap_remove_pages: paradoxical hold count:"
					" %d, va: 0x%x, pte: %d, pdirindex: %d\n",
				mpte->hold_count, pv->pv_va, mpte->pindex, pdirindex);
		}
#endif

		tpte = *pte;

#if defined(DIAGNOSTIC)
		/*
		 * Should have a page table entry for every pv entry
		 */
		if (tpte == NULL) {
			printf("pmap_remove_pages: missing mapping:"
					" va: 0x%x\n", pv->pv_va);
			continue;
		}
#endif

/*
 * We cannot remove wired pages from a process' mapping at this time
 */
		if (tpte & PG_W) {
			continue;
		}

		ppv = pa_to_pvh(tpte);

		setpte(pte, 0);

		pmap->pm_stats.resident_count--;
		m = ppv->pv_vm_page;

		/*
		 * Update the vm_page_t clean and reference bits.
		 */
		if (pv->pv_flags & PV_TRACKMOD) {
			if (tpte & PG_M) {
				m->dirty = VM_PAGE_BITS_ALL;
			}
			if (tpte & PG_A) {
				m->flags |= PG_REFERENCED;
			}
		}

		TAILQ_REMOVE(&pmap->pm_pvlist, pv, pv_plist);

		ppv->pv_list_count--;
		TAILQ_REMOVE(&ppv->pv_list, pv, pv_list);

		if (TAILQ_FIRST(&ppv->pv_list) == NULL) {
			m->flags &= ~(PG_MAPPED | PG_WRITEABLE);
		}

		if (pmap_unuse_pt(pmap, pv->pv_va, mpte)) {
			mpte = NULL;
		}

		free_pv_entry(pv);
	}

	page_wakeupq(&mpte);

	splx(s);
#if defined(SMP)
	invltlbcpu(pmap->pm_pactive);
#else
	invltlb();
#endif

}

/*
 * pmap_testbit tests bits in pte's
 * note that the testbit/changebit routines are inline,
 * and a lot of things compile-time evaluate.
 */
static __inline boolean_t
pmap_testbit(m, bit)
	vm_page_t m;
	int bit;
{
	pv_entry_t pv;
	pv_table_t *ppv;
	unsigned *pte;
	int s;

	if (m->flags & PG_FICTITIOUS)
		return FALSE;

	ppv = m->pventry;
	if (!ppv)
		return FALSE;

	s = splvm();
	/*
	 * Not found, check current mappings returning immediately if found.
	 */
	for (pv = TAILQ_FIRST(&ppv->pv_list);
	     pv;
	     pv = TAILQ_NEXT(pv, pv_list)) {
		/*
		 * if the bit being tested is the modified bit, then
		 * mark clean_map and ptes as never
		 * modified.
		 */
		if ((pv->pv_flags & PV_TRACKMOD) == 0)
			continue;

#if defined(PMAP_DIAGNOSTIC)
		if (pv->pv_pmap == NULL) {
			printf("Null pmap (tb) at va: 0x%x\n", pv->pv_va);
			continue;
		}
#endif
		pte = pmap_pte(pv->pv_pmap, pv->pv_va);
		if (pte && (*pte & bit)) {
			splx(s);
			return TRUE;
		}
	}
	splx(s);
	return (FALSE);
}

/*
 * this routine is used to modify bits in ptes
 */
static void
pmap_changebit(m, bit, setem)
	vm_page_t m;
	int bit;
	boolean_t setem;
{
	pmap_t pmap;
	pv_entry_t pv;
	pv_table_t *ppv;
	register unsigned *pte;
	unsigned pbits;
	int update_needed;
	int s;

	if (m->flags & PG_FICTITIOUS)
		return;

	ppv = m->pventry;
	if (!ppv)
		return;

	s = splvm();
	update_needed = 0;
	/*
	 * Loop over all current mappings setting/clearing as appropos If
	 * setting RO do we need to clear the VAC?
	 */
	for (pv = TAILQ_FIRST(&ppv->pv_list);
	     pv;
	     pv = TAILQ_NEXT(pv, pv_list)) {

#if defined(PMAP_DIAGNOSTIC)
		if (pv->pv_pmap == NULL) {
			printf("Null pmap (cb) at va: 0x%x\n", pv->pv_va);
			continue;
		}
#endif

		pte = pmap_pte(pv->pv_pmap, pv->pv_va);
		if (pte == NULL)
			continue;

		pbits = *(unsigned *)pte;

		if (setem) {
			pbits |= bit;
		} else {
			pbits &= ~bit;
		}

		if (pv->pv_flags & PV_TRACKMOD) {
			if (pbits & PG_A) {
				m->flags |= PG_REFERENCED;
				pbits &= ~PG_A;
			}

			if (pbits & PG_M) {
				m->dirty = VM_PAGE_BITS_ALL;
				pbits &= ~PG_M;
			}
		}

		if (*(unsigned *)pte != pbits) {
			update_needed |= pv->pv_pmap->pm_pactive;
			setpte(pte, pbits);
		}
	}
	splx(s);
#if defined(SMP)
	if (update_needed)
		invltlbcpu(update_needed);
#else
	if (update_needed)
		invltlb();
#endif
}

/*
 *      pmap_page_protect:
 *
 *      Lower the permission for all mappings to a given page.
 */
void
pmap_page_protect(vm_page_t m, vm_prot_t prot)
{
	if ((prot & VM_PROT_WRITE) == 0) {
		if (prot & (VM_PROT_READ | VM_PROT_EXECUTE)) {
			pmap_changebit(m, PG_RW, FALSE);
		} else {
			pmap_remove_all(m);
		}
	}
}

vm_offset_t
pmap_phys_address(ppn)
	int ppn;
{
	return (i386_ptob(ppn));
}

/*
 *	pmap_ts_referenced:
 *
 *	Return the count of reference bits for a page, clearing all of them.
 *	
 */
int
pmap_ts_referenced(vm_page_t m)
{
	register pv_entry_t pv, pvf, npv;
	pv_table_t *ppv;
	unsigned *pte, tpte;
	int s;
	int update_needed;
	int rtval;

	if (m->flags & PG_FICTITIOUS)
		return 0;

	ppv = m->pventry;
	if (!ppv)
		return 0;

	s = splvm();

	update_needed = 0;
	rtval = 0;

	pv = TAILQ_FIRST(&ppv->pv_list);
	pvf = pv;
	/*
	 * Not found, check current mappings returning immediately if found.
	 */
	for (; pv; pv = npv) {

		npv = TAILQ_NEXT(pv, pv_list);
		if (npv == pvf)
			npv = NULL;
		TAILQ_REMOVE(&ppv->pv_list, pv, pv_list);
		TAILQ_INSERT_TAIL(&ppv->pv_list, pv, pv_list);

		/*
		 * if the bit being tested is the modified bit, then
		 * mark clean_map and ptes as never
		 * modified.
		 */
		if ((pv->pv_flags & PV_TRACKMOD) == 0) {
			continue;
		}

		pte = pmap_pte(pv->pv_pmap, pv->pv_va);
		if (pte == NULL) {
			continue;
		}

		if ((tpte = *pte) & PG_A) {
			rtval++;
			update_needed |= pv->pv_pmap->pm_pactive;
			setpte(pte, (tpte & ~PG_A));
		}
	}

	splx(s);

#if defined(SMP)
	if (update_needed)
		invltlbcpu(update_needed);
#else
	if (update_needed)
		invltlb();
#endif
	return (rtval);
}

/*
 *	pmap_is_modified:
 *
 *	Return whether or not the specified physical page was modified
 *	in any physical maps.
 */
boolean_t
pmap_is_modified(vm_page_t m)
{
	boolean_t rtval;
	rtval = pmap_testbit(m, PG_M);
#if 0
	if (rtval) {
		m->dirty = VM_PAGE_BITS_ALL;
	}
#endif
	return rtval;
}

/*
 *	Clear the modify bits on the specified physical page.
 */
void
pmap_clear_modify(vm_page_t m)
{
	pmap_changebit(m, PG_M, FALSE);
}

/*
 *	pmap_clear_reference:
 *
 *	Clear the reference bit on the specified physical page.
 */
void
pmap_clear_reference(vm_page_t m)
{
	pmap_changebit(m, PG_A, FALSE);
}

/*
 * Miscellaneous support routines follow
 */

static void
i386_protection_init()
{
	register int *kp, prot;

	kp = protection_codes;
	for (prot = 0; prot < 8; prot++) {
		switch (prot) {
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_NONE:
			/*
			 * Read access is also 0. There isn't any execute bit,
			 * so just make it readable.
			 */
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_NONE:
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_EXECUTE:
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_EXECUTE:
			*kp++ = 0;
			break;
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_NONE:
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_EXECUTE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_NONE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE:
			*kp++ = PG_RW;
			break;
		}
	}
}

/*
 * Map a set of physical memory pages into the kernel virtual
 * address space. Return a pointer to where it is mapped. This
 * routine is intended to be used for mapping device memory,
 * NOT real memory.
 */
void *
pmap_mapdev(pa, size)
	vm_offset_t pa;
	vm_size_t size;
{
	vm_offset_t va, tmpva;
	unsigned *pte;

	size = roundup(size, PAGE_SIZE);

	va = kmem_alloc_pageable(kernel_map, size);
#if !defined(MAX_PERF)
	if (!va)
		panic("pmap_mapdev: Couldn't alloc kernel virtual memory");
#endif

	pa = pa & PG_FRAME;
	for (tmpva = va; size > 0;) {
		pte = (unsigned *)vtopte(tmpva);
		setpte(pte, pa | PG_RW | PG_M | PG_A | PG_V | staticpgeflag);
		size -= PAGE_SIZE;
		tmpva += PAGE_SIZE;
		pa += PAGE_SIZE;
	}
	invltlb();

	return ((void *) va);
}

/*
 * perform the pmap work for mincore
 */
int
pmap_mincore(pmap, addr)
	pmap_t pmap;
	vm_offset_t addr;
{
	
	unsigned *ptep, pte;
	vm_page_t m;
	int val = 0;
	
	ptep = pmap_pte(pmap, addr);
	if (ptep == 0) {
		return 0;
	}

	if (pte = *ptep) {
		pv_table_t *ppv;
		vm_offset_t pa;

		val = MINCORE_INCORE;
		if ((pte & PG_MANAGED) == 0)
			return val;

		pa = pte & PG_FRAME;

		ppv = pa_to_pvh((pa & PG_FRAME));
		m = ppv->pv_vm_page;

		/*
		 * Modified by us
		 */
		if (pte & PG_M)
			val |= MINCORE_MODIFIED|MINCORE_MODIFIED_OTHER;
		/*
		 * Modified by someone
		 */
		else if (m->dirty || pmap_is_modified(m))
			val |= MINCORE_MODIFIED_OTHER;
		/*
		 * Referenced by us
		 */
		if (pte & PG_A)
			val |= MINCORE_REFERENCED|MINCORE_REFERENCED_OTHER;

		/*
		 * Referenced by someone
		 */
		else if ((m->flags & PG_REFERENCED) || pmap_ts_referenced(m)) {
			val |= MINCORE_REFERENCED_OTHER;
			m->flags |= PG_REFERENCED;
		}
	} 
	return val;
}

void
pmap_activate(struct proc *p)
{
#if defined(SWTCH_OPTIM_STATS)
	tlb_flush_count++;
#endif
	load_cr3(p->p_addr->u_pcb.pcb_cr3 =
		vtophys(p->p_vmspace->vm_pmap.pm_pdir));
}

vm_offset_t
pmap_addr_hint(vm_object_t obj, vm_offset_t addr, vm_size_t size) {

	if ((obj == NULL) || (size < NBPDR) || (obj->type != OBJT_DEVICE)) {
		return addr;
	}

	addr = (addr + (NBPDR - 1)) & ~(NBPDR - 1);
	return addr;
}


#if defined(PMAP_DEBUG)
pmap_pid_dump(int pid) {
	pmap_t pmap;
	struct proc *p;
	int npte = 0;
	int index;
	for (p = allproc.lh_first; p != NULL; p = p->p_list.le_next) {
		if (p->p_pid != pid)
			continue;

		if (p->p_vmspace) {
			int i,j;
			index = 0;
			pmap = &p->p_vmspace->vm_pmap;
			for(i=0;i<1024;i++) {
				pd_entry_t *pde;
				unsigned *pte;
				unsigned base = i << PDRSHIFT;
				
				pde = &pmap->pm_pdir[i];
				if (pde && pmap_pde_v(pde)) {
					for(j=0;j<1024;j++) {
						unsigned va = base + (j << PAGE_SHIFT);
						if (va >= (vm_offset_t) VM_MIN_KERNEL_ADDRESS) {
							if (index) {
								index = 0;
								printf("\n");
							}
							return npte;
						}
						pte = pmap_pte( pmap, va);
						if (pte && pmap_pte_v(pte)) {
							vm_offset_t pa;
							vm_page_t m;
							pa = *(int *)pte;
							m = PHYS_TO_VM_PAGE((pa & PG_FRAME));
							printf("va: 0x%x, pt: 0x%x, h: %d, w: %d, f: 0x%x",
								va, pa, m->hold_count, m->wire_count, m->flags);
							npte++;
							index++;
							if (index >= 2) {
								index = 0;
								printf("\n");
							} else {
								printf(" ");
							}
						}
					}
				}
			}
		}
	}
	return npte;
}
#endif

#if defined(DEBUG)

static void	pads __P((pmap_t pm));
static void	pmap_pvdump __P((vm_offset_t pa));

/* print address space of pmap*/
static void
pads(pm)
	pmap_t pm;
{
	unsigned va, i, j;
	unsigned *ptep;

	if (pm == kernel_pmap)
		return;
	for (i = 0; i < 1024; i++)
		if (pm->pm_pdir[i])
			for (j = 0; j < 1024; j++) {
				va = (i << PDRSHIFT) + (j << PAGE_SHIFT);
				if (pm == kernel_pmap && va < KERNBASE)
					continue;
				if (pm != kernel_pmap && va > UPT_MAX_ADDRESS)
					continue;
				ptep = pmap_pte(pm, va);
				if (pmap_pte_v(ptep))
					printf("%x:%x ", va, *(int *) ptep);
			};

}

static void
pmap_pvdump(pa)
	vm_offset_t pa;
{
	pv_table_t *ppv;
	register pv_entry_t pv;

	printf("pa %x", pa);
	ppv = pa_to_pvh(pa);
	for (pv = TAILQ_FIRST(&ppv->pv_list);
		pv;
		pv = TAILQ_NEXT(pv, pv_list)) {
#ifdef used_to_be
		printf(" -> pmap %p, va %x, flags %x",
		    (void *)pv->pv_pmap, pv->pv_va, pv->pv_flags);
#endif
		printf(" -> pmap %p, va %x", (void *)pv->pv_pmap, pv->pv_va);
		pads(pv->pv_pmap);
	}
	printf(" ");
}
#endif
